#!/usr/bin/env python
# Copyright 2017-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: James D. McClain
#          Mario Motta
#          Yang Gao
#          Qiming Sun <osirpt.sun@gmail.com>
#          Jason Yu
#


import itertools
import numpy as np

from pyscf import lib
from pyscf.lib import logger
from pyscf.pbc.cc import eom_kccsd_ghf as eom_kgccsd
from pyscf.pbc.cc import kccsd
from pyscf.pbc.lib import kpts_helper
from pyscf.lib.parameters import LOOSE_ZERO_TOL, LARGE_DENOM  # noqa
from pyscf.pbc.cc import kintermediates_uhf
from pyscf.pbc.mp.kump2 import (get_frozen_mask, get_nocc, get_nmo,
                                padded_mo_coeff, padding_k_idx)  # noqa

einsum = lib.einsum

########################################
# EOM-IP-CCSD
########################################

def amplitudes_to_vector_ip(r1, r2, kshift, kconserv):
    r1a, r1b = r1
    r2aaa, r2baa, r2abb, r2bbb = r2
    nkpts = r2aaa.shape[0]
    nocca, noccb = r1a.shape[0], r1b.shape[0]
    nvira, nvirb = r2aaa.shape[-1], r2bbb.shape[-1]
    # From symmetry for aaa and bbb terms, only store lower
    # triangular part (ki,i) < (kj,j)
    idxa, idya = np.tril_indices(nkpts*nocca, -1)
    idxb, idyb = np.tril_indices(nkpts*noccb, -1)
    r2aaa = r2aaa.transpose(0,2,1,3,4).reshape(nkpts*nocca,nkpts*nocca,nvira)
    r2bbb = r2bbb.transpose(0,2,1,3,4).reshape(nkpts*noccb,nkpts*noccb,nvirb)
    return np.hstack((r1a, r1b, r2aaa[idxa,idya].ravel(),
                      r2baa.ravel(), r2abb.ravel(),
                      r2bbb[idxb,idyb].ravel()))

def vector_to_amplitudes_ip(vector, kshift, nkpts, nmo, nocc, kconserv):
    nocca, noccb = nocc
    nmoa, nmob = nmo
    nvira, nvirb = nmoa-nocca, nmob-noccb

    sizes = (nocca, noccb, (nkpts*nocca)*(nkpts*nocca-1)*nvira//2,
             nkpts**2*noccb*nocca*nvira, nkpts**2*nocca*noccb*nvirb,
             nkpts*noccb*(nkpts*noccb-1)*nvirb//2)
    sections = np.cumsum(sizes[:-1])
    r1a, r1b, r2a, r2baa, r2abb, r2b = np.split(vector, sections)

    r2a = r2a.reshape(nkpts*nocca*(nkpts*nocca-1)//2,nvira)
    r2b = r2b.reshape(nkpts*noccb*(nkpts*noccb-1)//2,nvirb)

    idxa, idya = np.tril_indices(nkpts*nocca, -1)
    idxb, idyb = np.tril_indices(nkpts*noccb, -1)

    r2aaa = np.zeros((nkpts*nocca,nkpts*nocca,nvira), dtype=r2a.dtype)
    r2aaa[idxa,idya] = r2a.copy()
    r2aaa[idya,idxa] = -r2a.copy()  # Fill in value :  kj, j < ki, i
    r2aaa = r2aaa.reshape(nkpts,nocca,nkpts,nocca,nvira)
    r2aaa = r2aaa.transpose(0,2,1,3,4)
    r2baa = r2baa.reshape(nkpts,nkpts,noccb,nocca,nvira).copy()
    r2abb = r2abb.reshape(nkpts,nkpts,nocca,noccb,nvirb).copy()
    r2bbb = np.zeros((nkpts*noccb,nkpts*noccb,nvirb), dtype=r2b.dtype)
    r2bbb[idxb,idyb] = r2b.copy()
    r2bbb[idyb,idxb] = -r2b.copy()  # Fill in value :  kj, j < ki, i
    r2bbb = r2bbb.reshape(nkpts,noccb,nkpts,noccb,nvirb)
    r2bbb = r2bbb.transpose(0,2,1,3,4)

    r1 = (r1a.copy(), r1b.copy())
    r2 = (r2aaa, r2baa, r2abb, r2bbb)
    return r1, r2

def ipccsd_matvec(eom, vector, kshift, imds=None, diag=None):
    '''2ph operators are of the form s_{ij}^{ b}, i.e. 'jb' indices are coupled'''
    if imds is None: imds = eom.make_imds()
    t1, t2= imds.t1, imds.t2
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    nocca, noccb, nvira, nvirb = t2ab.shape[3:]
    nmoa, nmob = nocca + nvira, noccb + nvirb
    kconserv = imds.kconserv
    nkpts = eom.nkpts

    r1, r2 = eom.vector_to_amplitudes(vector, kshift, nkpts, (nmoa, nmob), (nocca, noccb), kconserv)




    #nocc = eom.nocc
    #nmo = eom.nmo
    #nvir = (nmo[0]-nocc[0], nmo[1]-nocc[1])
    #nocca, noccb = nocc
    #nvira, nvirb = nvir
    #nkpts = eom.nkpts
    #r1, r2 = eom.vector_to_amplitudes(vector, nkpts, nmo[0]+nmo[1], nocc[0]+nocc[1])  # spin
    #spatial_r1, spatial_r2 = eom_kgccsd.spin2spatial_ip_doublet(r1, r2, kconserv, kshift, orbspin)
    #imds = imds._imds
    #t2aa, t2ab, t2bb = t2

    # k-point spin orbital version of ipccsd

    #Hr1 = -0.0*np.einsum('mi,m->i', imds.Foo[kshift], r1)

    #Hr2 = np.zeros_like(r2)

    r1a, r1b = r1
    r2aaa, r2baa, r2abb, r2bbb = r2

    #Foo term
    # -\sum_{kk,k} U_{kk,k,ki,i} s_{kk,k}
    Hr1a = -np.einsum('mi,m->i', imds.Foo[kshift], r1a)
    Hr1b = -np.einsum('MI,M->I', imds.FOO[kshift], r1b)

    #Fov term
    # \sum_{kL,kD,L,D} U_{kL,kD,L,D} S_{ki,i,kL,L}^{kD,D} + \sum_{kl,kd,l,d} U_{kl,kd,l,d} S_{ki,i,kl,l}^{kd,d}
    for km in range(nkpts):
        Hr1a += einsum('me,mie->i', imds.Fov[km], r2aaa[km,kshift])
        Hr1a -= einsum('ME,iME->i', imds.FOV[km], r2abb[kshift,km])
        Hr1b += einsum('ME,MIE->I', imds.FOV[km], r2bbb[km,kshift])
        Hr1b -= einsum('me,Ime->I', imds.Fov[km], r2baa[kshift,km])

    #Wooov
    # \sum_{kk,kl,kd,k,l,d} W_{kk,ki,kl,kd,k,i,l,d} s_{kl,kk,l,k}^{kd,d}
    # \sum_{kk,kL,kD,k,L,D} W_{kk,ki,kL,kD,k,i,L,D} s_{kL,kk,L,k}^{kD,D}
    for km in range(nkpts):
        for kn in range(nkpts):
            Hr1a += -0.5 * einsum('nime,mne->i', imds.Wooov[kn,kshift,km], r2aaa[km,kn])
            Hr1b +=        einsum('NIme,Nme->I', imds.WOOov[kn,kshift,km], r2baa[kn,km])
            Hr1b += -0.5 * einsum('NIME,MNE->I', imds.WOOOV[kn,kshift,km], r2bbb[km,kn])
            Hr1a +=        einsum('niME,nME->i', imds.WooOV[kn,kshift,km], r2abb[kn,km])

    dtype = np.result_type(Hr1a, *r2)
    Hr2aaa = np.zeros((nkpts, nkpts, nocca, nocca, nvira), dtype=dtype)
    Hr2baa = np.zeros((nkpts, nkpts, noccb, nocca, nvira), dtype=dtype)
    Hr2abb = np.zeros((nkpts, nkpts, nocca, noccb, nvirb), dtype=dtype)
    Hr2bbb = np.zeros((nkpts, nkpts, noccb, noccb, nvirb), dtype=dtype)

    # Fvv term
    # \sum_{kd,d} U_{kb,kd,b,d} S_{ki,kj,i,j}^{kd,d} = (\bar{H}S)_{ki,kj,i,j}^{kb,b}
    # \sum_{kD,D} S_{ki,kJ,i,J}^{kD,D} U_{kB,kD,B,D} = (\bar{H}S)_{ki,kJ,i,J}^{kB,B}
    for kb, ki in itertools.product(range(nkpts),repeat=2):
        kj = kconserv[kshift,ki,kb]
        Hr2aaa[ki,kj] += lib.einsum('be,ije->ijb', imds.Fvv[kb], r2aaa[ki,kj])
        Hr2abb[ki,kj] += lib.einsum('BE,iJE->iJB', imds.FVV[kb], r2abb[ki,kj])
        Hr2bbb[ki,kj] += lib.einsum('BE,IJE->IJB', imds.FVV[kb], r2bbb[ki,kj])
        Hr2baa[ki,kj] += lib.einsum('be,Ije->Ijb', imds.Fvv[kb], r2baa[ki,kj])

    # Foo term
    # \sum_{kl,l} U_{kl,ki,l,i} s_{kl,kj,l,j}^{kb,b} = (\bar{H}S)_{ki,kj,i,j}^{kb,b}
    # \sum_{kl,l} U_{kl,kj,l,j} S_{ki,kl,i,l}^{kb,b} = (\bar{H}S)_{ki,kj,i,j}^{kb,b}

    # \sum_{kl,l} S_{kl,kJ,l,J}^{kB,B} U_{kl,ki,l,i} = (\bar{H}S)_{ki,kJ,i,J}^{kB,B}
    # \sum_{KL,L} S_{ki,kL,i,L}^{kB,B} U_{kL,kJ,L,J} = (\bar{H}S)_{ki,kJ,i,J}^{kB,B}
    for ki, kj in itertools.product(range(nkpts), repeat=2):
        tmpa = lib.einsum('mi,mjb->ijb', imds.Foo[ki], r2aaa[ki,kj])
        tmpb = lib.einsum('mj,mib->ijb', imds.Foo[kj], r2aaa[kj,ki])
        Hr2aaa[ki,kj] -= tmpa - tmpb
        Hr2abb[ki,kj] -= lib.einsum('mi,mJB->iJB', imds.Foo[ki], r2abb[ki,kj])
        Hr2abb[ki,kj] -= lib.einsum('MJ,iMB->iJB', imds.FOO[kj], r2abb[ki,kj])
        Hr2baa[ki,kj] -= lib.einsum('MI,Mjb->Ijb', imds.FOO[ki], r2baa[ki,kj])
        Hr2baa[ki,kj] -= lib.einsum('mj,Imb->Ijb', imds.Foo[kj], r2baa[ki,kj])
        tmpb = lib.einsum('MI,MJB->IJB', imds.FOO[ki], r2bbb[ki,kj])
        tmpa = lib.einsum('MJ,MIB->IJB', imds.FOO[kj], r2bbb[kj,ki])
        Hr2bbb[ki,kj] -= tmpb - tmpa

    # Wovoo term
    # \sum_{kk,k} W_{kk,kb,kj,ki,k,b,j,i} s_{kk,k} = (\bar{H}S)_{ki,kj,i,j}^{kb,b}
    # \sum_{kk,k} W_{kk,kB,ki,kJ,k,B,i,J} S_{kk,k} = (\bar{H}S)_{ki,kJ,i,J}^{kB,B}
    for ki, kj in itertools.product(range(nkpts), repeat=2):
        kb = kconserv[ki, kshift, kj]
        Hr2aaa[ki,kj] -= einsum('mjbi,m->ijb', imds.Woovo[kshift,kj,kb], r1a)
        Hr2abb[ki,kj] += einsum('miBJ,m->iJB', imds.WooVO[kshift,ki,kb], r1a)
        Hr2baa[ki,kj] += einsum('MIbj,M->Ijb', imds.WOOvo[kshift,ki,kb], r1b)
        Hr2bbb[ki,kj] -= einsum('MJBI,M->IJB', imds.WOOVO[kshift,kj,kb], r1b)

    # Woooo term
    # \sum_{kk,kl,k,l} W_{kk,ki,kl,kj,k,i,l,j} S_{kk,kl,k,l}^{kb,b} = (\bar{H}S)_{ki,kj,i,j}^{kb,b}
    # \sum_{kk,kL,k,L} W_{kk,kL,ki,kJ,k,L,i,J} S_{kk,kl,k,L}^{kB,B} = (\bar{H}S)_{ki,kJ,i,J}^{kB,B}
    for ki, kj in itertools.product(range(nkpts), repeat=2):
        kb = kconserv[ki, kshift, kj]
        for kn in range(nkpts):
            km = kconserv[kj, kn, ki]
            Hr2aaa[ki, kj] += .5 * lib.einsum('minj,mnb->ijb', imds.Woooo[km, ki, kn], r2aaa[km, kn])
            Hr2abb[ki, kj] +=      lib.einsum('miNJ,mNB->iJB', imds.WooOO[km, ki, kn], r2abb[km, kn])
            Hr2bbb[ki, kj] += .5 * lib.einsum('MINJ,MNB->IJB', imds.WOOOO[km, ki, kn], r2bbb[km, kn])
            Hr2baa[ki, kj] +=      lib.einsum('njMI,Mnb->Ijb', imds.WooOO[kn, kj, km], r2baa[km, kn])

    # T2 term
    # - \sum_{kc,c} t_{kj,ki,j,i}^{kb,kc,b,c} [ \sum_{kk,kL,kD,k,L,D} W_{kL,kk,kD,kc,L,k,D,c} S_{kk,kL,k,L}^{kD,D}
    # + \sum{kk,kl,kd,k,l,d} W_{kl,kk,kd,kc,l,k,d,c} S_{kk,kl,k,l}^{kd,d} ] = (\bar{H}S)_{ki,kj,i,j}^{kb,b}
    #
    # - \sum_{kc,c} t_{ki,kJ,i,J}^{kc,kB,c,B} [ \sum_{kk,kL,kD,k,L,D} W_{kL,kk,kD,kc,L,k,D,c} S_{Kk,kL,k,L}^{kD,D}
    # + \sum{kk,kl,kd,k,l,d} W_{kl,kk,kd,kc,l,k,d,c} S_{kk,kl,k,l}^{kd,d} ] = (\bar{H}S)_{ki,kJ,i,J}^{kB,B}

    tmp_aaa = lib.einsum('xymenf,xymnf->e', imds.Wovov[:,kshift,:], r2aaa)
    tmp_bbb = lib.einsum('xyMENF,xyMNF->E', imds.WOVOV[:,kshift,:], r2bbb)
    tmp_abb = lib.einsum('xymeNF,xymNF->e', imds.WovOV[:,kshift,:], r2abb)
    tmp_baa = np.zeros(tmp_bbb.shape, dtype=tmp_bbb.dtype)
    for km, kn in itertools.product(range(nkpts), repeat=2):
        kf = kconserv[kn, kshift, km]
        tmp_baa += lib.einsum('nfME, Mnf->E', imds.WovOV[kn, kf, km], r2baa[km, kn])


    for ki, kj in itertools.product(range(nkpts), repeat=2):
        kb = kconserv[ki, kshift, kj]

        Hr2aaa[ki,kj] -= 0.5 * lib.einsum('e,jibe->ijb', tmp_aaa, t2aa[kj,ki,kb])
        Hr2aaa[ki,kj] -= lib.einsum('e,jibe->ijb', tmp_abb, t2aa[kj,ki,kb])

        Hr2abb[ki,kj] -= 0.5 * lib.einsum('e,iJeB->iJB', tmp_aaa, t2ab[ki,kj,kshift])
        Hr2abb[ki,kj] -= lib.einsum('e,iJeB->iJB', tmp_abb, t2ab[ki,kj,kshift])

        Hr2baa[ki,kj] -= 0.5 * lib.einsum('E,jIbE->Ijb', tmp_bbb, t2ab[kj,ki,kb])
        Hr2baa[ki,kj] -= lib.einsum('E,jIbE->Ijb', tmp_baa, t2ab[kj,ki,kb])

        Hr2bbb[ki,kj] -= 0.5 * lib.einsum('E,JIBE->IJB', tmp_bbb, t2bb[kj,ki,kb])
        Hr2bbb[ki,kj] -= lib.einsum('E,JIBE->IJB', tmp_baa, t2bb[kj,ki,kb])

    #idxoa = [np.where(orbspin[k][:nocca+noccb] == 0)[0] for k in range(nkpts)]
    #idxva = [np.where(orbspin[k][nocca+noccb:] == 0)[0] for k in range(nkpts)]
    #idxob = [np.where(orbspin[k][:nocca+noccb] == 1)[0] for k in range(nkpts)]
    #idxvb = [np.where(orbspin[k][nocca+noccb:] == 1)[0] for k in range(nkpts)]

    # j \/ b   |  i
    #    ---   |
    #      /\  |
    #    m \/ e|
    #     -------
    for ki, kj in itertools.product(range(nkpts), repeat=2):
        kb = kconserv[ki, kshift, kj]
        for km in range(nkpts):
            ke = kconserv[km, kshift, ki]

            # \sum_{kL,kD,L,D} W_{kL,kD,kb,kj,L,D,b,j} S_{ki,kL,i,L}^{kb,b}
            # \sum_{kl,kd,l,d} W_{kl,kd,kb,kj,l,d,b,j} S_{ki,kl,i,l}^{kb,b}
            Hr2aaa[ki, kj] += lib.einsum('mebj,ime->ijb', imds.Wovvo[km, ke, kb],
                                         r2aaa[ki, km])
            Hr2aaa[ki, kj] += lib.einsum('MEbj,iME->ijb', imds.WOVvo[km, ke, kb],
                                         r2abb[ki, km])
            # P(ij)
            ke = kconserv[km, kshift, kj]
            Hr2aaa[ki, kj] -= lib.einsum('mebi,jme->ijb', imds.Wovvo[km, ke, kb],
                                         r2aaa[kj, km])
            Hr2aaa[ki, kj] -= lib.einsum('MEbi,jME->ijb', imds.WOVvo[km, ke, kb],
                                         r2abb[kj, km])

            # \sum_{kL,kD,L,D} W_{kL,kD,kb,kJ,L,D,b,J} S_{ki,kL,i,L}^{kD,D}
            # \sum_{kl,kd,l,d} W_{kl,kd,kB,kJ,l,d,B,J} S_{ki,kl,i,l}^{kd,d}
            ke = kconserv[km, kshift, ki]
            Hr2abb[ki, kj] += lib.einsum('meBJ,ime->iJB', imds.WovVO[km, ke, kb],
                                         r2aaa[ki, km])
            Hr2abb[ki, kj] += lib.einsum('MEBJ,iME->iJB', imds.WOVVO[km, ke, kb],
                                         r2abb[ki, km])
            ke = kconserv[km, kshift, kj]
            Hr2abb[ki, kj] -= lib.einsum('miBE,mJE->iJB', imds.WooVV[km, ki, kb],
                                         r2abb[km, kj])

            ke = kconserv[km, kshift, ki]
            Hr2baa[ki, kj] += lib.einsum('MEbj,IME->Ijb', imds.WOVvo[km, ke, kb],
                                         r2bbb[ki, km])
            Hr2baa[ki, kj] += lib.einsum('mebj,Ime->Ijb', imds.Wovvo[km, ke, kb],
                                         r2baa[ki, km])
            ke = kconserv[km, kshift, kj]
            Hr2baa[ki, kj] -= lib.einsum('MIbe,Mje->Ijb', imds.WOOvv[km, ki, kb],
                                         r2baa[km, kj])

            ke = kconserv[km, kshift, ki]
            Hr2bbb[ki, kj] += lib.einsum('MEBJ,IME->IJB', imds.WOVVO[km, ke, kb],
                                         r2bbb[ki, km])
            Hr2bbb[ki, kj] += lib.einsum('meBJ,Ime->IJB', imds.WovVO[km, ke, kb],
                                         r2baa[ki, km])
            # P(ij)
            ke = kconserv[km, kshift, kj]
            Hr2bbb[ki, kj] -= lib.einsum('MEBI,JME->IJB', imds.WOVVO[km, ke, kb],
                                         r2bbb[kj, km])
            Hr2bbb[ki, kj] -= lib.einsum('meBI,Jme->IJB', imds.WovVO[km, ke, kb],
                                         r2baa[kj, km])

    #spatial_Hr1 = [Hr1a, Hr1b]
    #spatial_Hr2 = [Hr2aaa, Hr2baa, Hr2abb, Hr2bbb]
    #spin_Hr1, spin_Hr2 = eom_kgccsd.spatial2spin_ip_doublet(spatial_Hr1, spatial_Hr2,
#                                                            kconserv, kshift, orbspin)
    #Hr1 += spin_Hr1
    #Hr2 += spin_Hr2
    #vector = eom.amplitudes_to_vector(Hr1, Hr2)
    vector = amplitudes_to_vector_ip([Hr1a, Hr1b], [Hr2aaa, Hr2baa, Hr2abb, Hr2bbb], kshift, kconserv)
    return vector

def ipccsd_diag(eom, kshift, imds=None):
    if imds is None: imds = eom.make_imds()
    t1, t2 = imds.t1, imds.t2
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    nkpts, nocc_a, nvir_a = t1a.shape
    nkpts, nocc_b, nvir_b = t1b.shape
    kconserv = imds.kconserv

    Hr1a = -np.diag(imds.Foo[kshift])
    Hr1b = -np.diag(imds.FOO[kshift])

    Hr2aaa = np.zeros((nkpts,nkpts,nocc_a,nocc_a,nvir_a), dtype=t1[0].dtype)
    Hr2bbb = np.zeros((nkpts,nkpts,nocc_b,nocc_b,nvir_b), dtype=t1[0].dtype)
    Hr2abb = np.zeros((nkpts,nkpts,nocc_a,nocc_b,nvir_b), dtype=t1[0].dtype)
    Hr2baa = np.zeros((nkpts,nkpts,nocc_b,nocc_a,nvir_a), dtype=t1[0].dtype)
    if eom.partition == 'mp':
        raise Exception("MP diag is not tested") # remove this to use untested code
        #foo = eris.fock[0][:,:nocc_a,:nocc_a]
        #fOO = eris.fock[1][:,:nocc_b,:nocc_b]
        #fvv = eris.fock[0][:,:nvir_a,:nvir_a]
        #fVV = eris.fock[1][:,:nvir_b,:nvir_b]
        for ki in range(nkpts):
            for kj in range(nkpts):
                ka = kconserv[ki,kshift,kj]
                Hr2aaa[ki,kj]  = imds.Fvv[ka].diagonal()
                Hr2aaa[ki,kj] -= imds.Foo[ki].diagonal()[:,None,None]
                Hr2aaa[ki,kj] -= imds.Foo[kj].diagonal()[None,:,None]
                Hr2bbb[ki,kj]  = imds.FVV[ka].diagonal()
                Hr2bbb[ki,kj] -= imds.FOO[ki].diagonal()[:,None,None]
                Hr2bbb[ki,kj] -= imds.FOO[kj].diagonal()[None,:,None]
                Hr2aba[ki,kj]  = imds.Fvv[ka].diagonal()
                Hr2aba[ki,kj] -= imds.Foo[ki].diagonal()[:,None,None]
                Hr2aba[ki,kj] -= imds.FOO[kj].diagonal()[None,:,None]
                Hr2bab[ki,kj]  = imds.FVV[ka].diagonal()
                Hr2bab[ki,kj] -= imds.FOO[ki].diagonal()[:,None,None]
                Hr2bab[ki,kj] -= imds.Foo[kj].diagonal()[None,:,None]
    else:
        for ka in range(nkpts):
            for ki in range(nkpts):
                kj = kconserv[kshift,ki,ka]
                Hr2aaa[ki,kj] += imds.Fvv[ka].diagonal()
                Hr2abb[ki,kj] += imds.FVV[ka].diagonal()
                Hr2bbb[ki,kj] += imds.FVV[ka].diagonal()
                Hr2baa[ki,kj] += imds.Fvv[ka].diagonal()

                Hr2aaa[ki,kj] -= imds.Foo[ki].diagonal()[:,None,None]
                Hr2aaa[ki,kj] -= imds.Foo[kj].diagonal()[None,:,None]
                Hr2abb[ki,kj] -= imds.Foo[ki].diagonal()[:,None,None]
                Hr2abb[ki,kj] -= imds.FOO[kj].diagonal()[None,:,None]
                Hr2baa[ki,kj] -= imds.FOO[ki].diagonal()[:,None,None]
                Hr2baa[ki,kj] -= imds.Foo[kj].diagonal()[None,:,None]
                Hr2bbb[ki,kj] -= imds.FOO[ki].diagonal()[:,None,None]
                Hr2bbb[ki,kj] -= imds.FOO[kj].diagonal()[None,:,None]

        for ki, kj in itertools.product(range(nkpts), repeat=2):
            Hr2aaa[ki, kj] += lib.einsum('iijj->ij', imds.Woooo[ki, ki, kj])[:,:,None]
            Hr2abb[ki, kj] += lib.einsum('iiJJ->iJ', imds.WooOO[ki, ki, kj])[:,:,None]
            Hr2bbb[ki, kj] += lib.einsum('IIJJ->IJ', imds.WOOOO[ki, ki, kj])[:,:,None]
            Hr2baa[ki, kj] += lib.einsum('jjII->Ij', imds.WooOO[kj, kj, ki])[:,:,None]

            kb = kconserv[ki, kshift, kj]
            Hr2aaa[ki,kj] -= lib.einsum('iejb,jibe->ijb', imds.Wovov[ki,kshift,kj], t2aa[kj,ki,kb])
            Hr2abb[ki,kj] -= lib.einsum('ieJB,iJeB->iJB', imds.WovOV[ki,kshift,kj], t2ab[ki,kj,kshift])
            Hr2baa[ki,kj] -= lib.einsum('jbIE,jIbE->Ijb', imds.WovOV[kj,kb,ki], t2ab[kj,ki,kb])
            Hr2bbb[ki,kj] -= lib.einsum('IEJB,JIBE->IJB', imds.WOVOV[ki,kshift,kj], t2bb[kj,ki,kb])

            Hr2aaa[ki, kj] += lib.einsum('ibbi->ib', imds.Wovvo[ki, kb, kb])[:,None,:]
            Hr2aaa[ki, kj] += lib.einsum('jbbj->jb', imds.Wovvo[kj, kb, kb])[None,:,:]

            Hr2baa[ki, kj] += lib.einsum('jbbj->jb', imds.Wovvo[kj, kb, kb])[None,:,:]
            Hr2baa[ki, kj] -= lib.einsum('IIbb->Ib', imds.WOOvv[ki, ki, kb])[:,None,:]

            Hr2abb[ki, kj] += lib.einsum('JBBJ->JB', imds.WOVVO[kj, kb, kb])[None,:,:]
            Hr2abb[ki, kj] -= lib.einsum('iiBB->iB', imds.WooVV[ki, ki, kb])[:,None,:]

            Hr2bbb[ki, kj] += lib.einsum('IBBI->IB', imds.WOVVO[ki, kb, kb])[:,None,:]
            Hr2bbb[ki, kj] += lib.einsum('JBBJ->JB', imds.WOVVO[kj, kb, kb])[None,:,:]

    vector = amplitudes_to_vector_ip((Hr1a,Hr1b), (Hr2aaa,Hr2baa,Hr2abb,Hr2bbb), kshift, kconserv)
    return vector

def mask_frozen_ip(eom, vector, kshift, const=LARGE_DENOM):
    '''Replaces all frozen orbital indices of `vector` with the value `const`.'''
    nkpts = eom.nkpts
    nocca, noccb = eom.nocc
    nmoa, nmob = eom.nmo
    kconserv = eom.kconserv

    r1, r2 = eom.vector_to_amplitudes(vector, kshift, nkpts, (nmoa, nmob), (nocca, noccb), kconserv)
    r1a, r1b = r1
    r2aaa, r2baa, r2abb, r2bbb = r2

    # Get location of padded elements in occupied and virtual space
    nonzero_opadding, nonzero_vpadding = eom.nonzero_opadding, eom.nonzero_vpadding
    nonzero_opadding_a, nonzero_opadding_b = nonzero_opadding
    nonzero_vpadding_a, nonzero_vpadding_b = nonzero_vpadding

    new_r1a = const * np.ones_like(r1a)
    new_r1b = const * np.ones_like(r1b)
    new_r2aaa = const * np.ones_like(r2aaa)
    new_r2baa = const * np.ones_like(r2baa)
    new_r2abb = const * np.ones_like(r2abb)
    new_r2bbb = const * np.ones_like(r2bbb)

    # r1a/b case
    new_r1a[nonzero_opadding_a[kshift]] = r1a[nonzero_opadding_a[kshift]]
    new_r1b[nonzero_opadding_b[kshift]] = r1b[nonzero_opadding_b[kshift]]

    # r2aaa case
    for ki in range(nkpts):
        for kj in range(nkpts):
            kb = kconserv[ki, kshift, kj]
            idx = np.ix_([ki], [kj], nonzero_opadding_a[ki], nonzero_opadding_a[kj], nonzero_vpadding_a[kb])
            new_r2aaa[idx] = r2aaa[idx]

    # r2baa case
    for ki in range(nkpts):
        for kj in range(nkpts):
            kb = kconserv[ki, kshift, kj]
            idx = np.ix_([ki], [kj], nonzero_opadding_b[ki], nonzero_opadding_a[kj], nonzero_vpadding_a[kb])
            new_r2baa[idx] = r2baa[idx]

    # r2abb case
    for ki in range(nkpts):
        for kj in range(nkpts):
            kb = kconserv[ki, kshift, kj]
            idx = np.ix_([ki], [kj], nonzero_opadding_a[ki], nonzero_opadding_b[kj], nonzero_vpadding_b[kb])
            new_r2abb[idx] = r2abb[idx]

    # r2bbb case
    for ki in range(nkpts):
        for kj in range(nkpts):
            kb = kconserv[ki, kshift, kj]
            idx = np.ix_([ki], [kj], nonzero_opadding_b[ki], nonzero_opadding_b[kj], nonzero_vpadding_b[kb])
            new_r2bbb[idx] = r2bbb[idx]

    return eom.amplitudes_to_vector((new_r1a,new_r1b), (new_r2aaa,new_r2baa,new_r2abb,new_r2bbb), kshift, kconserv)


def get_padding_k_idx(eom, cc):
    # Get location of padded elements in occupied and virtual space
    nonzero_padding_alpha, nonzero_padding_beta = padding_k_idx(cc, kind="split")
    nonzero_opadding_alpha, nonzero_vpadding_alpha = nonzero_padding_alpha
    nonzero_opadding_beta, nonzero_vpadding_beta = nonzero_padding_beta
    return ((nonzero_opadding_alpha, nonzero_opadding_beta),
            (nonzero_vpadding_alpha, nonzero_vpadding_beta))


class EOMIP(eom_kgccsd.EOMIP):
    def __init__(self, cc):
        #if not isinstance(cc, kccsd.GCCSD):
        #    raise TypeError
        self.kpts = cc.kpts
        eom_kgccsd.EOMIP.__init__(self, cc)

    get_diag = ipccsd_diag
    matvec = ipccsd_matvec
    get_padding_k_idx = get_padding_k_idx
    mask_frozen = mask_frozen_ip

    def get_init_guess(self, kshift, nroots=1, koopmans=False, diag=None):
        size = self.vector_size()
        dtype = getattr(diag, 'dtype', np.complex128)
        nroots = min(nroots, size)
        nocca, noccb = self.nocc
        guess = []
        if koopmans:
            idx = np.zeros(nroots, dtype=int)
            tmp_oalpha, tmp_obeta = self.nonzero_opadding[kshift]
            tmp_oalpha = list(tmp_oalpha)
            tmp_obeta = list(tmp_obeta)
            if len(tmp_obeta) + len(tmp_oalpha) < nroots:
                raise ValueError("Max number of roots for k-point (idx=%3d) for koopmans "
                                 "is %3d.\nRequested %3d." %
                                 (kshift, len(tmp_obeta)+len(tmp_oalpha), nroots))

            total_count = 0
            while (total_count < nroots):
                if total_count % 2 == 0 and len(tmp_oalpha) > 0:
                    idx[total_count] = tmp_oalpha.pop()
                else:
                    # Careful! index depends on how we create vector
                    # (here the first elements are r1a, then r1b)
                    idx[total_count] = nocca + tmp_obeta.pop()
                total_count += 1
        else:
            idx = diag.argsort()

        for i in idx[:nroots]:
            g = np.zeros(size, dtype)
            g[i] = 1.0
            g = self.mask_frozen(g, kshift, const=0.0)
            guess.append(g)
        return guess

    def gen_matvec(self, kshift, imds=None, left=False, **kwargs):
        if imds is None: imds = self.make_imds()
        diag = self.get_diag(kshift, imds)
        if left:
            raise NotImplementedError
            matvec = lambda xs: [self.l_matvec(x, kshift, imds, diag) for x in xs]
        else:
            matvec = lambda xs: [self.matvec(x, kshift, imds, diag) for x in xs]
        return matvec, diag

    def vector_to_amplitudes(self, vector, kshift, nkpts=None, nmo=None, nocc=None, kconserv=None):
        if nmo is None: nmo = self.nmo
        if nocc is None: nocc = self.nocc
        if nkpts is None: nkpts = self.nkpts
        if kconserv is None: kconserv = self.kconserv
        return vector_to_amplitudes_ip(vector, kshift, nkpts, nmo, nocc, kconserv)

    def amplitudes_to_vector(self, r1, r2, kshift, kconserv=None):
        if kconserv is None: kconserv = self.kconserv
        return amplitudes_to_vector_ip(r1, r2, kshift, kconserv)

    def vector_size(self):
        nocca, noccb = self.nocc
        nmoa, nmob = self.nmo
        nvira, nvirb = nmoa - nocca, nmob - noccb
        nkpts = self.nkpts
        return (nocca + noccb +
                nkpts*nocca*(nkpts*nocca-1)*nvira//2 +
                nkpts**2*noccb*nocca*nvira +
                nkpts**2*nocca*noccb*nvirb +
                nkpts*noccb*(nkpts*noccb-1)*nvirb//2)

    def make_imds(self, eris=None, t1=None, t2=None):
        imds = _IMDS(self._cc, eris, t1, t2)
        imds.make_ip()
        return imds


########################################
# EOM-EA-CCSD
########################################

def amplitudes_to_vector_ea(r1, r2, kshift, kconserv):
    r1a, r1b = r1
    r2a, r2aba, r2bab, r2b = r2
    nkpts = r2a.shape[0]
    nocca, noccb = r2a.shape[2], r2b.shape[2]
    nvira, nvirb = r2a.shape[3], r2b.shape[3]
    # From symmetry for aaa and bbb terms, only store lower
    # triangular part (ka,a) < (kb,b)
    r2aaa = np.zeros((nocca*nkpts*nvira*(nkpts*nvira-1))//2, dtype=r2a.dtype)
    r2bbb = np.zeros((noccb*nkpts*nvirb*(nkpts*nvirb-1))//2, dtype=r2b.dtype)

    index = 0
    for kj, ka in itertools.product(range(nkpts), repeat=2):
        kb = kconserv[kshift,ka,kj]
        if ka < kb:  # Take diagonal part
            idxa, idya = np.tril_indices(nvira, 0)
        else:  # Don't take diagonal (equal to zero)
            idxa, idya = np.tril_indices(nvira, -1)
        r2aaa[index:index + nocca*len(idya)] = r2a[kj,ka,:,idxa,idya].reshape(-1)
        index = index + nocca*len(idya)

    index = 0
    for kj, ka in itertools.product(range(nkpts), repeat=2):
        kb = kconserv[kshift,ka,kj]
        if ka < kb:  # Take diagonal part
            idxb, idyb = np.tril_indices(nvirb, 0)
        else:
            idxb, idyb = np.tril_indices(nvirb, -1)
        r2bbb[index:index + noccb*len(idyb)] = r2b[kj,ka,:,idxb,idyb].reshape(-1)
        index = index + noccb*len(idyb)

    return np.hstack((r1a, r1b, r2aaa.ravel(),
                      r2aba.ravel(), r2bab.ravel(),
                      r2bbb.ravel()))

def vector_to_amplitudes_ea(vector, kshift, nkpts, nmo, nocc, kconserv):
    nocca, noccb = nocc
    nmoa, nmob = nmo
    nvira, nvirb = nmoa-nocca, nmob-noccb

    sizes = (nvira, nvirb, nkpts*nocca*(nkpts*nvira-1)*nvira//2,
             nkpts**2*nocca*nvirb*nvira, nkpts**2*noccb*nvira*nvirb,
             nkpts*noccb*(nkpts*nvirb-1)*nvirb//2)
    sections = np.cumsum(sizes[:-1])
    r1a, r1b, r2a, r2aba, r2bab, r2b = np.split(vector, sections)

    r2aaa = np.zeros((nkpts,nkpts,nocca,nvira,nvira), dtype=r2a.dtype)
    r2aba = r2aba.reshape(nkpts,nkpts,nocca,nvirb,nvira).copy()
    r2bab = r2bab.reshape(nkpts,nkpts,noccb,nvira,nvirb).copy()
    r2bbb = np.zeros((nkpts,nkpts,noccb,nvirb,nvirb), dtype=r2b.dtype)

    index = 0
    for kj, ka in itertools.product(range(nkpts), repeat=2):
        kb = kconserv[kshift,ka,kj]
        if ka < kb:  # Take diagonal part
            idxa, idya = np.tril_indices(nvira, 0)
        else:
            idxa, idya = np.tril_indices(nvira, -1)
        tmp = r2a[index:index + nocca*len(idya)].reshape(-1,nocca)
        r2aaa[kj,ka,:,idxa,idya] = tmp
        r2aaa[kj,kb,:,idya,idxa] = -tmp
        index = index + nocca*len(idya)

    index = 0
    for kj, ka in itertools.product(range(nkpts), repeat=2):
        kb = kconserv[kshift,ka,kj]
        if ka < kb:  # Take diagonal part
            idxb, idyb = np.tril_indices(nvirb, 0)
        else:
            idxb, idyb = np.tril_indices(nvirb, -1)
        tmp = r2b[index:index + noccb*len(idyb)].reshape(-1,noccb)
        r2bbb[kj,ka,:,idxb,idyb] = tmp
        r2bbb[kj,kb,:,idyb,idxb] = -tmp
        index = index + noccb*len(idyb)

    r1 = (r1a.copy(), r1b.copy())
    r2 = (r2aaa, r2aba, r2bab, r2bbb)
    return r1, r2

def eaccsd_matvec(eom, vector, kshift, imds=None, diag=None):
    '''2ph operators are of the form s_{ j}^{ab}, i.e. 'jb' indices are coupled'''
    if imds is None: imds = eom.make_imds()
    t1, t2= imds.t1, imds.t2
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    nocca, noccb, nvira, nvirb = t2ab.shape[3:]
    nmoa, nmob = nocca + nvira, noccb + nvirb
    kconserv = imds.kconserv
    nkpts = eom.nkpts

    r1, r2 = eom.vector_to_amplitudes(vector, kshift, nkpts, (nmoa, nmob), (nocca, noccb), kconserv)

    r1a, r1b = r1
    r2aaa, r2aba, r2bab, r2bbb = r2


    # BEGINNING OF MATVEC CONTRACTIONS: ref - Nooijen 1995 EOM-CC for EA

    # Fvv terms
    # (\bar{H}S)^a = \sum_{kc,c} U_{ac} s^c
    Hr1a = einsum('ac,c->a', imds.Fvv[kshift], r1a)
    Hr1b = einsum('AC,C->A', imds.FVV[kshift], r1b)

    # Fov terms
    # (\bar{H}S)^a = \sum_{kL,kD, L, D} U_{kL,kD,L,D} s^{a,kD,D}_{kL,L} + \sum_{kl,kd,l,d} U_{kl, d}^{a,kd,d}_{kl,l}
    for kl in range(nkpts):
        Hr1a += einsum('ld,lad->a', imds.Fov[kl], r2aaa[kl,kshift])
        Hr1a += einsum('LD,LaD->a', imds.FOV[kl], r2bab[kl,kshift])
        Hr1b += einsum('ld,lAd->A', imds.Fov[kl], r2aba[kl,kshift])
        Hr1b += einsum('LD,LAD->A', imds.FOV[kl], r2bbb[kl,kshift])

    # Wvovv
    # (\bar{H}S)^a = \sum_{kc,kL,kD,c,L,D} W_{kL,kc,kD,a,l,c,D} s_{kL,L}^{kc,kD,c,D}
    # + \sum_{kc,kd,kl,c,d,l} W_{ka,kl,kc,kd,a,l,c,d} s_{kl,l}^{kc,kd,c,d}
    for kc, kl in itertools.product(range(nkpts), repeat=2):
        Hr1a += 0.5*lib.einsum('acld,lcd->a', imds.Wvvov[kshift,kc,kl], r2aaa[kl,kc])
        Hr1a +=     lib.einsum('acLD,LcD->a', imds.WvvOV[kshift,kc,kl], r2bab[kl,kc])
        Hr1b += 0.5*lib.einsum('ACLD,LCD->A', imds.WVVOV[kshift,kc,kl], r2bbb[kl,kc])
        Hr1b +=     lib.einsum('ACld,lCd->A', imds.WVVov[kshift,kc,kl], r2aba[kl,kc])

    dtype = np.result_type(Hr1a, *r2)
    Hr2aaa = np.zeros((nkpts, nkpts, nocca, nvira, nvira), dtype=dtype)
    Hr2aba = np.zeros((nkpts, nkpts, nocca, nvirb, nvira), dtype=dtype)
    Hr2bab = np.zeros((nkpts, nkpts, noccb, nvira, nvirb), dtype=dtype)
    Hr2bbb = np.zeros((nkpts, nkpts, noccb, nvirb, nvirb), dtype=dtype)

    # Wvvvv
    # \sum_{kc,kd,c,d} W_{ka,kb,kc,kd,a,b,c,d} s_{kj,j}^{kc,kd,c,d} = (\bar{H}S)^{kb, a, b}_{kj,j}
    # \sum_{kc,kD,c,D} W{ka,kB,kc,kD,a,B,c,D} s_{kJ,kc,kD,J,c,D} = (\bar{H}S)^{kB, a, B}_{kJ,J}
    for kj, ka in itertools.product(range(nkpts), repeat=2):
        kb = kconserv[kshift,ka,kj]
        for kc in range(nkpts):
            kd = kconserv[ka, kc, kb]
            Wvvvv, WvvVV, WVVVV = imds.get_Wvvvv(ka, kb, kc)
            Hr2aaa[kj,ka] += .5 * lib.einsum('acbd,jcd->jab', Wvvvv, r2aaa[kj,kc])
            Hr2aba[kj,kb] +=      lib.einsum('bcad,jdc->jab', WvvVV, r2aba[kj,kd])
            Hr2bab[kj,ka] +=      lib.einsum('acbd,jcd->jab', WvvVV, r2bab[kj,kc])
            Hr2bbb[kj,ka] += .5 * lib.einsum('acbd,jcd->jab', WVVVV, r2bbb[kj,kc])

    #Wvvvo
    # \sum_{kc,ka,kj,c,a,j} W_{kb,kc,kj,a,b,c,j} s^{kc,c} = (\bar{H}S)^{kb, a, b}_{kj,j}
    # \sum_{kc,ka,kJ,c,a,J} W_{kB,kc,kJ,a,B,c,J} s^{kc,c} = (\bar{H}S)^{kB, a, B}_{kJ,J}
    for ka, kj, in itertools.product(range(nkpts),repeat=2):
        kb = kconserv[kshift,ka,kj]
        kc = kshift
        Hr2aaa[kj,ka] += einsum('acbj,c->jab', imds.Wvvvo[ka,kc,kb], r1a)
        Hr2bbb[kj,ka] += einsum('ACBJ,C->JAB', imds.WVVVO[ka,kc,kb], r1b)

        Hr2bab[kj,ka] += einsum('acBJ,c->JaB', imds.WvvVO[ka,kc,kb], r1a)
        Hr2aba[kj,ka] += einsum('ACbj,C->jAb', imds.WVVvo[ka,kc,kb], r1b)

    #Fvv Terms
    # sum_{kc,ka,kj,c,a,j} s_{kj,j}^{kc,kb,c,b} U_{ka,kc,a,c} = (\bar{H}S)^{kb, a, b}_{kj,j}
    # sum_{kd,ka,kj,d,b,j} s_{kj,j}^{ka,kd,a,d} U_{kb,kd,b,d} = (\bar{H}S)^{kb, a, b}_{kj,j}

    # sum_{kc,ka,kJ,c,a,J} U_{ka,kc,a,c} s_{kJ,J}^{kc,kB,c,B} = (\bar{H}S)^{kB, a, B}_{kJ,J}
    # sum_{kD,ka,kj,D,a,j} U_{kb,kd,b,d} s_{kj,j}^{ka,kd,a,d} = (\bar{H}S)^{kB, a, B}_{kJ,J}
    for ka, kj in itertools.product(range(nkpts), repeat=2):
        # kb = kshift - ka + kj
        kb = kconserv[kshift, ka, kj]
        tmpa = lib.einsum('ac,jcb->jab', imds.Fvv[ka], r2aaa[kj,ka])
        tmpb = lib.einsum('bc,jca->jab', imds.Fvv[kb], r2aaa[kj,kb])
        Hr2aaa[kj,ka] += tmpa - tmpb
        Hr2aba[kj,ka] += lib.einsum('AC,jCb->jAb', imds.FVV[ka], r2aba[kj,ka])
        Hr2bab[kj,ka] += lib.einsum('ac,JcB->JaB', imds.Fvv[ka], r2bab[kj,ka])
        Hr2aba[kj,ka] += lib.einsum('bc, jAc -> jAb', imds.Fvv[kb], r2aba[kj,ka])
        Hr2bab[kj,ka] += lib.einsum('BC, JaC -> JaB', imds.FVV[kb], r2bab[kj,ka])
        tmpb = lib.einsum('AC,JCB->JAB', imds.FVV[ka], r2bbb[kj,ka])
        tmpa = lib.einsum('BC,JCA->JAB', imds.FVV[kb], r2bbb[kj,kb])
        Hr2bbb[kj,ka] += tmpb - tmpa

    #Foo Term
    # \sum_{ka,kl,l} U_{kl,l,kj,j} s^{ka,a,kb,b}^{kl,l} = (\bar{H}S)^{kb, a, b}_{kj,j}
    # \sum_{ka,kL,L} s^{ka,a,kB,B}_{kL,L} U_{kL,L,kJ,J} = (\bar{H}S)^{kB, a, B}_{kJ,J}
    for kl, ka in itertools.product(range(nkpts), repeat=2):
        Hr2aaa[kl,ka] -= lib.einsum('lj,lab->jab', imds.Foo[kl], r2aaa[kl,ka])
        Hr2bbb[kl,ka] -= lib.einsum('LJ,LAB->JAB', imds.FOO[kl], r2bbb[kl,ka])
        Hr2bab[kl,ka] -= lib.einsum('LJ,LaB->JaB', imds.FOO[kl], r2bab[kl,ka])
        Hr2aba[kl,ka] -= lib.einsum('lj,lAb->jAb', imds.Foo[kl], r2aba[kl,ka])

    # Woovv term
    # - \sum{kk,k} t_{kk,kj,k,j}^{ka,kb,a,b} [\sum_{kc,kD,kL,c,D,L} W_{kL,kk,kD,kc,L,k,D,c} s_{kL,L}^{kc,kD,c,D}
    # + \sum{kc,kd,kl,c,d,l} W_{kk,kl,kc,kd,k,l,c,d} s_{kl,l}^{kc,kd,c,d} ] = (\bar{H}S)^{kb, a, b}_{kj,j}
    #
    # - \sum_{kk,k} t_{kk,kJ,k,J}^{ka,kB,a,B} [ \sum{kc,kD,kL,c,D,L} W_{kk,kL,kc,kD,k,L,c,D} s_{kL,L}^{kc,kD,c,D}
    # + \sum_{kc,kd,kl,c,d,l} W_{kk,kl,kc,kd,k,l,c,d} s_{kl,l}^{kc,kd,c,d} ] = (\bar{H}S)^{kB, a, B}_{kJ,J}
    tmp_aaa = lib.einsum('xykcld, yxlcd->k', imds.Wovov[kshift,:,:], r2aaa)
    tmp_bbb = lib.einsum('xyKCLD, yxLCD->K', imds.WOVOV[kshift,:,:], r2bbb)
    tmp_bab = lib.einsum('xykcLD, yxLcD->k', imds.WovOV[kshift], r2bab)
    tmp_aba = np.zeros(tmp_bbb.shape, dtype = tmp_bbb.dtype)

    for kl, kc in itertools.product(range(nkpts), repeat=2):
        kd = kconserv[kl,kc,kshift]
        tmp_aba += lib.einsum('ldKC, lCd->K', imds.WovOV[kl,kd,kshift], r2aba[kl,kc])

    Hr2aaa -= 0.5 * lib.einsum('k, xykjab->xyjab', tmp_aaa, t2aa[kshift])
    Hr2bab -= 0.5 * lib.einsum('k, xykJaB->xyJaB', tmp_aaa, t2ab[kshift])

    Hr2aaa -= lib.einsum('k, xykjab->xyjab', tmp_bab, t2aa[kshift])
    Hr2bbb -= 0.5 * lib.einsum('K, xyKJAB->xyJAB', tmp_bbb, t2bb[kshift])

    Hr2bbb -= lib.einsum('K, xyKJAB->xyJAB', tmp_aba, t2bb[kshift])
    Hr2bab -= lib.einsum('k, xykJaB->xyJaB', tmp_bab, t2ab[kshift])

    for kj, ka in itertools.product(range(nkpts), repeat=2):
        kb = kconserv[kshift, ka, kj]
        Hr2aba[kj, ka] -= lib.einsum('K, jKbA->jAb', tmp_aba, t2ab[kj, kshift, kb])
        Hr2aba[kj, ka] -= 0.5 * einsum('K, jKbA->jAb', tmp_bbb, t2ab[kj, kshift, kb])

    # j \/ b   |  a
    #    ---   |
    #      /\  |
    #    l \/ d|
    #     -------
    for kj, ka in itertools.product(range(nkpts), repeat=2):
        kb = kconserv[kshift, ka, kj]
        for kd in range(nkpts):
            kl = kconserv[ka, kshift, kd]

            # \sum_{kL,kD,L,D} W_{kL,kb,kD,kj,L,b,D,j} s_{kL,L}^{ka,kD,a,D} = (\bar{H}S)^{kb, a, b}_{kj,j}
            # \sum_{kl,kd,l,d} W_{kl,kb,kd,kj,l,b,d,j} s_{kl,l}^{ka,kd,a,d} = (\bar{H}S)^{kb, a, b}_{kj,j}
            Hr2aaa[kj, ka] += lib.einsum('ldbj,lad->jab', imds.Wovvo[kl,kd,kb],
                                         r2aaa[kl,ka])
            Hr2aaa[kj, ka] += lib.einsum('LDbj,LaD->jab', imds.WOVvo[kl,kd,kb],
                                         r2bab[kl,ka])
            # P(ab)
            kl = kconserv[kb, kshift, kd]
            Hr2aaa[kj, ka] -= lib.einsum('ldaj,lbd->jab', imds.Wovvo[kl,kd,ka],
                                         r2aaa[kl,kb])
            Hr2aaa[kj, ka] -= lib.einsum('LDaj,LbD->jab', imds.WOVvo[kl,kd,ka],
                                         r2bab[kl,kb])

            kl = kconserv[ka, kshift, kd]

            # \sum_{kL,kD,L,D} W_{kL,kB,kD,kJ,L,B,D,J} s_{kL,L}^{ka,kD,a,D} = (\bar{H}S)^{kB, a, B}_{kJ,J}
            # \sum_{kl,kd,l,d} W_{kl,kB,kd,kJ,l,B,d,J} s_{kl,l}^{ka,kd,a,d} = (\bar{H}S)^{kB, a, B}_{kJ,J}
            # - \sum_{kc,kL,c,L} W_{ka,kL,kc,kJ,a,L,c,J} s_{kL,L}^{kc,kB,c,B} = (\bar{H}S)^{kB, a, B}_{kJ,J}
            Hr2bab[kj, ka] += lib.einsum('ldBJ,lad->JaB', imds.WovVO[kl,kd,kb],
                                         r2aaa[kl,ka])
            Hr2bab[kj, ka] += lib.einsum('LDBJ,LaD->JaB', imds.WOVVO[kl,kd,kb],
                                         r2bab[kl,ka])
            kl = kconserv[kb, kshift, kd]
            Hr2bab[kj, ka] -= lib.einsum('LJad,LdB->JaB', imds.WOOvv[kl,kj,ka],
                                         r2bab[kl,kd])

            kl = kconserv[ka, kshift, kd]
            Hr2aba[kj, ka] += lib.einsum('LDbj,LAD->jAb', imds.WOVvo[kl,kd,kb],
                                         r2bbb[kl,ka])
            Hr2aba[kj, ka] += lib.einsum('ldbj,lAd->jAb', imds.Wovvo[kl,kd,kb],
                                         r2aba[kl,ka])
            kl = kconserv[kb, kshift, kd]
            Hr2aba[kj, ka] -= lib.einsum('ljAD,lDb->jAb', imds.WooVV[kl,kj,ka],
                                         r2aba[kl,kd])

            kl = kconserv[ka, kshift, kd]
            Hr2bbb[kj, ka] += lib.einsum('LDBJ,LAD->JAB', imds.WOVVO[kl,kd,kb],
                                         r2bbb[kl,ka])
            Hr2bbb[kj, ka] += lib.einsum('ldBJ,lAd->JAB', imds.WovVO[kl,kd,kb],
                                         r2aba[kl,ka])
            # P(ab)
            kl = kconserv[kb, kshift, kd]
            Hr2bbb[kj, ka] -= lib.einsum('LDAJ,LBD->JAB', imds.WOVVO[kl,kd,ka],
                                         r2bbb[kl,kb])
            Hr2bbb[kj, ka] -= lib.einsum('ldAJ,lBd->JAB', imds.WovVO[kl,kd,ka],
                                         r2aba[kl,kb])

    vector = amplitudes_to_vector_ea([Hr1a, Hr1b], [Hr2aaa, Hr2aba, Hr2bab, Hr2bbb], kshift, kconserv)
    return vector

def eaccsd_diag(eom, kshift, imds=None):
    if imds is None: imds = eom.make_imds()
    t1, t2 = imds.t1, imds.t2
    t1a, t1b = t1
    t2aa, t2ab, t2bb = t2
    nkpts, nocca, nvira = t1a.shape
    nkpts, noccb, nvirb = t1b.shape
    kconserv = imds.kconserv

    #Hr1a = np.zeros((nvira), dtype=t1a.dtype)
    #Hr1b = np.zeros((nvirb), dtype=t1b.dtype)
    Hr2aaa = np.zeros((nkpts,nkpts,nocca,nvira,nvira), dtype=t1a.dtype)
    Hr2aba = np.zeros((nkpts,nkpts,nocca,nvirb,nvira), dtype=t1a.dtype)
    Hr2bab = np.zeros((nkpts,nkpts,noccb,nvira,nvirb), dtype=t1a.dtype)
    Hr2bbb = np.zeros((nkpts,nkpts,noccb,nvirb,nvirb), dtype=t1b.dtype)

    Hr1a = np.diag(imds.Fvv[kshift])
    Hr1b = np.diag(imds.FVV[kshift])
    if eom.partition == 'mp':
        raise Exception("MP diag is not tested") # remove this to use untested code
        for kj, ka in itertools.product(range(nkpts), repeat=2):
            kb = kconserv[kshift, ka, kj]
            Hr2aaa[kj,ka] -= imds.Foo[kj,:,None,None]
            Hr2aaa[kj,ka] += imds.Fvv[ka,None,:,None]
            Hr2aaa[kj,ka] += imds.Fvv[kb,None,None,:]

            Hr2aba[kj,ka] -= imds.Foo[kj,:,None,None]
            Hr2aba[kj,ka] += imds.FVV[ka,None,:,None]
            Hr2aba[kj,ka] += imds.Fvv[kb,None,None,:]

            Hr2bab[kj,ka] -= imds.FOO[kj,:,None,None]
            Hr2bab[kj,ka] += imds.Fvv[ka,None,:,None]
            Hr2bab[kj,ka] += imds.FVV[kb,None,None,:]

            Hr2bbb[kj,ka] -= imds.FOO[kj,:,None,None]
            Hr2bbb[kj,ka] += imds.FVV[ka,None,:,None]
            Hr2bbb[kj,ka] += imds.FVV[kb,None,None,:]
    else:
        for kj, ka in itertools.product(range(nkpts), repeat=2):
            kb = kconserv[kshift, ka, kj]
            # Fvv
            Hr2aaa[kj,ka] += imds.Fvv[ka].diagonal()[None,:,None]
            Hr2aaa[kj,ka] += imds.Fvv[kb].diagonal()[None,None,:]
            Hr2aba[kj,ka] += imds.FVV[ka].diagonal()[None,:,None]
            Hr2aba[kj,ka] += imds.Fvv[kb].diagonal()[None,None,:]
            Hr2bab[kj,ka] += imds.Fvv[ka].diagonal()[None,:,None]
            Hr2bab[kj,ka] += imds.FVV[kb].diagonal()[None,None,:]
            Hr2bbb[kj,ka] += imds.FVV[ka].diagonal()[None,:,None]
            Hr2bbb[kj,ka] += imds.FVV[kb].diagonal()[None,None,:]

            # Foo
            Hr2aaa[kj,ka] -= imds.Foo[kj].diagonal()[:,None,None]
            Hr2bbb[kj,ka] -= imds.FOO[kj].diagonal()[:,None,None]
            Hr2bab[kj,ka] -= imds.FOO[kj].diagonal()[:,None,None]
            Hr2aba[kj,ka] -= imds.Foo[kj].diagonal()[:,None,None]

            # Wvvvv
            Wvvvv, WvvVV, WVVVV = imds.get_Wvvvv(ka, kb, ka)
            # FIXME: Do Wvvvv and WVVVV have a factor 0.5?
            Hr2aaa[kj,ka] += lib.einsum('aabb->ab', Wvvvv)[None,:,:]
            Hr2aba[kj,kb] += lib.einsum('bbAA->Ab', WvvVV)[None,:,:]
            Hr2bab[kj,ka] += lib.einsum('aaBB->aB', WvvVV)[None,:,:]
            Hr2bbb[kj,ka] += lib.einsum('AABB->AB', WVVVV)[None,:,:]

            # Wovov term (physicist's Woovv)
            Hr2aaa[kj,ka] -= lib.einsum('kajb, kjab->jab', imds.Wovov[kshift,ka,kj], t2aa[kshift,kj,ka])
            Hr2aba[kj,ka] -= lib.einsum('jbKA, jKbA->jAb', imds.WovOV[kj,kb,kshift], t2ab[kj,kshift,kb])
            Hr2bab[kj,ka] -= lib.einsum('kaJB, kJaB->JaB', imds.WovOV[kshift,ka,kj], t2ab[kshift,kj,ka])
            Hr2bbb[kj,ka] -= lib.einsum('kajb, kjab->jab', imds.WOVOV[kshift,ka,kj], t2bb[kshift,kj,ka])

            # Wovvo term
            Hr2aaa[kj, ka] += lib.einsum('jbbj->jb', imds.Wovvo[kj,kb,kb])[:,None,:]
            Hr2aaa[kj, ka] += lib.einsum('jaaj->ja', imds.Wovvo[kj,ka,ka])[:,:,None]

            Hr2aba[kj, ka] += lib.einsum('jbbj->jb', imds.Wovvo[kj,kb,kb])[:,None,:]
            Hr2aba[kj, ka] -= lib.einsum('jjAA->jA', imds.WooVV[kj,kj,ka])[:,:,None]

            Hr2bab[kj, ka] += lib.einsum('JBBJ->JB', imds.WOVVO[kj,kb,kb])[:,None,:]
            Hr2bab[kj, ka] -= lib.einsum('JJaa->Ja', imds.WOOvv[kj,kj,ka])[:,:,None]

            Hr2bbb[kj, ka] += lib.einsum('JBBJ->JB', imds.WOVVO[kj,kb,kb])[:,None,:]
            Hr2bbb[kj, ka] += lib.einsum('JAAJ->JA', imds.WOVVO[kj,ka,ka])[:,:,None]

    vector = amplitudes_to_vector_ea([Hr1a,Hr1b], [Hr2aaa,Hr2aba,Hr2bab,Hr2bbb], kshift, kconserv)
    return vector

def mask_frozen_ea(eom, vector, kshift, const=LARGE_DENOM):
    '''Replaces all frozen orbital indices of `vector` with the value `const`.'''
    nkpts = eom.nkpts
    nocca, noccb = eom.nocc
    nmoa, nmob = eom.nmo
    kconserv = eom.kconserv

    r1, r2 = eom.vector_to_amplitudes(vector, kshift, nkpts, (nmoa, nmob), (nocca, noccb), kconserv)
    r1a, r1b = r1
    r2aaa, r2aba, r2bab, r2bbb = r2

    # Get location of padded elements in occupied and virtual space
    nonzero_opadding, nonzero_vpadding = eom.nonzero_opadding, eom.nonzero_vpadding
    nonzero_opadding_a, nonzero_opadding_b = nonzero_opadding
    nonzero_vpadding_a, nonzero_vpadding_b = nonzero_vpadding

    new_r1a = const * np.ones_like(r1a)
    new_r1b = const * np.ones_like(r1b)
    new_r2aaa = const * np.ones_like(r2aaa)
    new_r2aba = const * np.ones_like(r2aba)
    new_r2bab = const * np.ones_like(r2bab)
    new_r2bbb = const * np.ones_like(r2bbb)

    # r1a/b case
    new_r1a[nonzero_vpadding_a[kshift]] = r1a[nonzero_vpadding_a[kshift]]
    new_r1b[nonzero_vpadding_b[kshift]] = r1b[nonzero_vpadding_b[kshift]]

    # r2aaa case
    for kj in range(nkpts):
        for ka in range(nkpts):
            kb = kconserv[kshift, ka, kj]
            idx = np.ix_([kj], [ka], nonzero_opadding_a[kj], nonzero_vpadding_a[ka], nonzero_vpadding_a[kb])
            new_r2aaa[idx] = r2aaa[idx]

    # r2aba case
    for kj in range(nkpts):
        for ka in range(nkpts):
            kb = kconserv[kshift, ka, kj]
            idx = np.ix_([kj], [ka], nonzero_opadding_a[kj], nonzero_vpadding_b[ka], nonzero_vpadding_a[kb])
            new_r2aba[idx] = r2aba[idx]

    # r2bab case
    for kj in range(nkpts):
        for ka in range(nkpts):
            kb = kconserv[kshift, ka, kj]
            idx = np.ix_([kj], [ka], nonzero_opadding_b[kj], nonzero_vpadding_a[ka], nonzero_vpadding_b[kb])
            new_r2bab[idx] = r2bab[idx]

    # r2bbb case
    for kj in range(nkpts):
        for ka in range(nkpts):
            kb = kconserv[kshift, ka, kj]
            idx = np.ix_([kj], [ka], nonzero_opadding_b[kj], nonzero_vpadding_b[ka], nonzero_vpadding_b[kb])
            new_r2bbb[idx] = r2bbb[idx]

    return eom.amplitudes_to_vector((new_r1a,new_r1b), (new_r2aaa,new_r2aba,new_r2bab,new_r2bbb), kshift)

class EOMEA(eom_kgccsd.EOMEA):
    def __init__(self, cc):
        #if not isinstance(cc, kccsd.GCCSD):
        #    raise TypeError
        self.kpts = cc.kpts
        eom_kgccsd.EOMEA.__init__(self, cc)

    get_diag = eaccsd_diag
    matvec = eaccsd_matvec
    get_padding_k_idx = get_padding_k_idx
    mask_frozen = mask_frozen_ea

    def get_init_guess(self, kshift, nroots=1, koopmans=False, diag=None):
        size = self.vector_size()
        dtype = getattr(diag, 'dtype', np.complex128)
        nroots = min(nroots, size)
        nocca, noccb = self.nocc
        nmoa, nmob = self.nmo
        nvira = nmoa-nocca
        guess = []
        if koopmans:
            idx = np.zeros(nroots, dtype=int)
            tmp_valpha, tmp_vbeta = self.nonzero_vpadding[kshift]
            tmp_valpha = list(tmp_valpha)
            tmp_vbeta = list(tmp_vbeta)
            if len(tmp_vbeta) + len(tmp_valpha) < nroots:
                raise ValueError("Max number of roots for k-point (idx=%3d) for koopmans "
                                 "is %3d.\nRequested %3d." %
                                 (kshift, len(tmp_vbeta)+len(tmp_valpha), nroots))

            total_count = 0
            while (total_count < nroots):
                if total_count % 2 == 0 and len(tmp_valpha) > 0:
                    idx[total_count] = tmp_valpha.pop(0)
                else:
                    # Careful! index depends on how we create vector
                    # (here the first elements are r1a, then r1b)
                    idx[total_count] = nvira + tmp_vbeta.pop(0)
                total_count += 1
        else:
            idx = diag.argsort()

        for i in idx[:nroots]:
            g = np.zeros(size, dtype)
            g[i] = 1.0
            g = self.mask_frozen(g, kshift, const=0.0)
            guess.append(g)
        return guess

    def vector_to_amplitudes(self, vector, kshift, nkpts=None, nmo=None, nocc=None, kconserv=None):
        if nmo is None: nmo = self.nmo
        if nocc is None: nocc = self.nocc
        if nkpts is None: nkpts = self.nkpts
        if kconserv is None: kconserv = self.kconserv
        return vector_to_amplitudes_ea(vector, kshift, nkpts, nmo, nocc, kconserv)

    def amplitudes_to_vector(self, r1, r2, kshift, kconserv=None):
        if kconserv is None: kconserv = self.kconserv
        return amplitudes_to_vector_ea(r1, r2, kshift, kconserv)

    def vector_size(self):
        nocca, noccb = self.nocc
        nmoa, nmob = self.nmo
        nvira, nvirb = nmoa - nocca, nmob - noccb
        nkpts = self.nkpts
        #return (nvira + nvirb +
        #        nocca*nkpts*nvira*nkpts*nvira +
        #        nkpts**2*nocca*nvirb*nvira +
        #        nkpts**2*noccb*nvira*nvirb +
        #        noccb*nkpts*nvirb*nkpts*nvirb)
        return (nvira + nvirb +
                nocca*nkpts*nvira*(nkpts*nvira-1)//2 +
                nkpts**2*nocca*nvirb*nvira +
                nkpts**2*noccb*nvira*nvirb +
                noccb*nkpts*nvirb*(nkpts*nvirb-1)//2)

    def make_imds(self, eris=None, t1=None, t2=None):
        imds = _IMDS(self._cc, eris, t1, t2)
        imds.make_ea()
        return imds

class _IMDS:
    def __init__(self, cc, eris=None, t1=None, t2=None):
        self._cc = cc
        self.verbose = cc.verbose
        self.kconserv = kpts_helper.get_kconserv(cc._scf.cell, cc.kpts)
        self.stdout = cc.stdout
        if t1 is None:
            t1 = cc.t1
        self.t1 = t1
        if t2 is None:
            t2 = cc.t2
        self.t2 = t2
        if eris is None:
            eris = cc.ao2mo()
        self.eris = eris
        self._made_shared = False
        self.made_ip_imds = False
        self.made_ea_imds = False
        self.made_ee_imds = False

    def _make_shared(self):
        cput0 = (logger.process_clock(), logger.perf_counter())

        t1, t2, eris = self.t1, self.t2, self.eris
        self.Foo, self.FOO = kintermediates_uhf.Foo(self._cc, t1, t2, eris)
        self.Fvv, self.FVV = kintermediates_uhf.Fvv(self._cc, t1, t2, eris)
        self.Fov, self.FOV = kintermediates_uhf.Fov(self._cc, t1, t2, eris)

        # 2 virtuals
        self.Wovvo, self.WovVO, self.WOVvo, self.WOVVO = kintermediates_uhf.Wovvo(self._cc, t1, t2, eris)
        self.Woovv, self.WooVV, self.WOOvv, self.WOOVV = kintermediates_uhf.Woovv(self._cc, t1, t2, eris)
        self.Wovov = eris.ovov - np.asarray(eris.ovov).transpose(2,1,0,5,4,3,6)
        self.WOVOV = eris.OVOV - np.asarray(eris.OVOV).transpose(2,1,0,5,4,3,6)
        self.WovOV = eris.ovOV
        self.WOVov = None

        self._made_shared = True
        logger.timer_debug1(self, 'EOM-KCCSD shared intermediates', *cput0)
        return self

    def make_ip(self):
        if not self._made_shared:
            self._make_shared()

        kconserv = self.kconserv
        cput0 = (logger.process_clock(), logger.perf_counter())

        t1, t2, eris = self.t1, self.t2, self.eris

        # 0 or 1 virtuals
        self.Woooo, self.WooOO, _         , self.WOOOO = kintermediates_uhf.Woooo(self._cc, t1, t2, eris)  # noqa: E501
        self.Wooov, self.WooOV, self.WOOov, self.WOOOV = kintermediates_uhf.Wooov(self._cc, t1, t2, eris, kconserv)  # noqa: E501
        self.Woovo, self.WooVO, self.WOOvo, self.WOOVO = kintermediates_uhf.Woovo(self._cc, t1, t2, eris)  # noqa: E501

        self.made_ip_imds = True
        logger.timer_debug1(self, 'EOM-KUCCSD IP intermediates', *cput0)
        return self

    def make_ea(self):
        if not self._made_shared:
            self._make_shared()

        cput0 = (logger.process_clock(), logger.perf_counter())

        t1, t2, eris = self.t1, self.t2, self.eris

        # 3 or 4 virtuals
        #self.Wvovv, self.WvoVV, self.WVOvv, self.WVOVV = kintermediates_uhf.Wvovv(self._cc, t1, t2, eris)
        self.Wvvov, self.WvvOV, self.WVVov, self.WVVOV = kintermediates_uhf.Wvvov(self._cc, t1, t2, eris)
        if eris.vvvv is not None:
            self.Wvvvv, self.WvvVV, self.WVVVV = kintermediates_uhf.Wvvvv(self._cc, t1, t2, eris)
        else:
            self.Wvvvv = self.WvvVV = self.WVVVV = None
        self.Wvvvo, self.WvvVO, self.WVVvo, self.WVVVO = kintermediates_uhf.Wvvvo(self._cc, t1, t2, eris)

        self.made_ea_imds = True
        logger.timer_debug1(self, 'EOM-KUCCSD EA intermediates', *cput0)
        return self

    def make_ee(self):
        raise NotImplementedError

    def get_Wvvvv(self, ka, kb, kc):
        if not self.made_ea_imds:
            self.make_ea()

        if self.Wvvvv is None:
            return kintermediates_uhf.get_Wvvvv(self._cc, self.t1, self.t2, self.eris,
                                                ka, kb, kc)
        else:
            return self.Wvvvv[ka,kc,kb], self.WvvVV[ka,kc,kb], self.WVVVV[ka,kc,kb]

if __name__ == '__main__':
    from pyscf.pbc import gto
    from pyscf.pbc import scf
    from pyscf import lo

    cell = gto.Cell()
    cell.atom='''
    He 0.000000000000   0.000000000000   0.000000000000
    He 1.685068664391   1.685068664391   1.685068664391
    '''
    #cell.basis = [[0, (1., 1.)], [1, (.5, 1.)]]
    cell.basis = [[0, (1., 1.)], [0, (.5, 1.)]]
    cell.a = '''
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000'''
    cell.unit = 'B'
    cell.mesh = [5, 5, 5]
    cell.build()

    np.random.seed(1)
    # Running HF and CCSD with 1x1x2 Monkhorst-Pack k-point mesh
    kmf = scf.KUHF(cell, kpts=cell.make_kpts([1,1,3]), exxdiv=None)
    nmo = cell.nao_nr()
    kmf.mo_occ = np.zeros((2,3,nmo))
    kmf.mo_occ[0,:,:3] = 1
    kmf.mo_occ[1,:,:1] = 1
    kmf.mo_energy = np.arange(nmo) + np.random.random((2,3,nmo)) * .3
    kmf.mo_energy[kmf.mo_occ == 0] += 2

    mo = (np.random.random((2,3,nmo,nmo)) +
          np.random.random((2,3,nmo,nmo))*1j - .5-.5j)
    s = kmf.get_ovlp()
    kmf.mo_coeff = np.empty_like(mo)
    nkpts = len(kmf.kpts)
    for k in range(nkpts):
        kmf.mo_coeff[0,k] = lo.orth.vec_lowdin(mo[0,k], s[k])
        kmf.mo_coeff[1,k] = lo.orth.vec_lowdin(mo[1,k], s[k])

    def rand_t1_t2(mycc):
        nkpts = mycc.nkpts
        nocca, noccb = mycc.nocc
        nmoa, nmob = mycc.nmo
        nvira, nvirb = nmoa - nocca, nmob - noccb
        np.random.seed(1)
        t1a = (np.random.random((nkpts,nocca,nvira)) +
               np.random.random((nkpts,nocca,nvira))*1j - .5-.5j)
        t1b = (np.random.random((nkpts,noccb,nvirb)) +
               np.random.random((nkpts,noccb,nvirb))*1j - .5-.5j)
        t2aa = (np.random.random((nkpts,nkpts,nkpts,nocca,nocca,nvira,nvira)) +
                np.random.random((nkpts,nkpts,nkpts,nocca,nocca,nvira,nvira))*1j - .5-.5j)
        kconserv = kpts_helper.get_kconserv(kmf.cell, kmf.kpts)
        t2aa = t2aa - t2aa.transpose(1,0,2,4,3,5,6)
        tmp = t2aa.copy()
        for ki, kj, kk in kpts_helper.loop_kkk(nkpts):
            kl = kconserv[ki, kk, kj]
            t2aa[ki,kj,kk] = t2aa[ki,kj,kk] - tmp[ki,kj,kl].transpose(0,1,3,2)
        t2ab = (np.random.random((nkpts,nkpts,nkpts,nocca,noccb,nvira,nvirb)) +
                np.random.random((nkpts,nkpts,nkpts,nocca,noccb,nvira,nvirb))*1j - .5-.5j)
        t2bb = (np.random.random((nkpts,nkpts,nkpts,noccb,noccb,nvirb,nvirb)) +
                np.random.random((nkpts,nkpts,nkpts,noccb,noccb,nvirb,nvirb))*1j - .5-.5j)
        t2bb = t2bb - t2bb.transpose(1,0,2,4,3,5,6)
        tmp = t2bb.copy()
        for ki, kj, kk in kpts_helper.loop_kkk(nkpts):
            kl = kconserv[ki, kk, kj]
            t2bb[ki,kj,kk] = t2bb[ki,kj,kk] - tmp[ki,kj,kl].transpose(0,1,3,2)

        t1 = (t1a, t1b)
        t2 = (t2aa, t2ab, t2bb)
        return t1, t2

    from pyscf.pbc.cc import kccsd_uhf
    mycc = kccsd_uhf.KUCCSD(kmf)
    eris = mycc.ao2mo()

    t1, t2 = rand_t1_t2(mycc)
    mycc.t1 = t1
    mycc.t2 = t2
    kconserv = kpts_helper.get_kconserv(kmf.cell, kmf.kpts)

    kgcc = kccsd.GCCSD(scf.addons.convert_to_ghf(kmf))
    kccsd_eris = kccsd._make_eris_incore(kgcc, kgcc._scf.mo_coeff)
    orbspin = kccsd_eris.orbspin
    nkpts = mycc.nkpts
    nocca, noccb = mycc.nocc
    nmoa, nmob = mycc.nmo
    nvira, nvirb = nmoa - nocca, nmob - noccb

    kshift = 0  # excitation out of 0th k-point
    nmo = nmoa + nmob
    nocc = nocca + noccb
    nvir = nmo - nocc

    np.random.seed(0)
    # IP version
    myeom = EOMIP(mycc)
    imds = myeom.make_imds()
    imds.make_ip()

    spin_r1_ip = (np.random.rand(nocc)*1j +
                  np.random.rand(nocc) - 0.5 - 0.5*1j)
    spin_r2_ip = (np.random.rand(nkpts**2 * nocc**2 * nvir) +
                  np.random.rand(nkpts**2 * nocc**2 * nvir)*1j - 0.5 - 0.5*1j)
    spin_r2_ip = spin_r2_ip.reshape(nkpts, nkpts, nocc, nocc, nvir)
    spin_r2_ip = eom_kgccsd.enforce_2p_spin_ip_doublet(spin_r2_ip, kconserv, kshift, orbspin)

    r1, r2 = eom_kgccsd.spin2spatial_ip_doublet(spin_r1_ip, spin_r2_ip, kconserv, kshift, orbspin)
    vector = myeom.amplitudes_to_vector(r1, r2, kshift)
    vector = myeom.matvec(vector, kshift=kshift, imds=imds)
    Hr1, Hr2 = myeom.vector_to_amplitudes(vector, nkpts, (nmoa, nmob), (nocca, noccb))
    Hr1a, Hr1b = Hr1
    Hr2aaa, Hr2baa, Hr2abb, Hr2bbb = Hr2
    print('ip Hr1a',   abs(lib.fp(Hr1a)   - (-0.34462696543560045-1.6104596956729178j)))
    print('ip Hr1b',   abs(lib.fp(Hr1b)   - (-0.055793611517250929+0.22169994342782473j)))
    print('ip Hr2aaa', abs(lib.fp(Hr2aaa) - (0.692705827672665420-1.958639508839846943j)))
    print('ip Hr2baa', abs(lib.fp(Hr2baa) - (2.892194153603884654+2.039530776282815872j)))
    print('ip Hr2abb', abs(lib.fp(Hr2abb) - (1.618257685489421727-5.489218743953674817j)))
    print('ip Hr2bbb', abs(lib.fp(Hr2bbb) - (0.479835513829048044+0.108406393138471210j)))
    # EA version

    myeom = EOMEA(mycc)
    imds = myeom.make_imds()
    imds.make_ea()

    spin_r1_ea = (np.random.rand(nvir)*1j +
                  np.random.rand(nvir) - 0.5 - 0.5*1j)
    spin_r2_ea = (np.random.rand(nkpts**2 * nocc * nvir**2) +
                  np.random.rand(nkpts**2 * nocc * nvir**2)*1j - 0.5 - 0.5*1j)
    spin_r2_ea = spin_r2_ea.reshape(nkpts, nkpts, nocc, nvir, nvir)
    spin_r2_ea = eom_kgccsd.enforce_2p_spin_ea_doublet(spin_r2_ea, kconserv, kshift, orbspin)
    r1, r2 = eom_kgccsd.spin2spatial_ea_doublet(spin_r1_ea, spin_r2_ea, kconserv, kshift, orbspin)

    vector = myeom.amplitudes_to_vector(r1, r2, kshift)
    vector = myeom.matvec(vector, kshift=kshift, imds=imds)
    Hr1, Hr2 = myeom.vector_to_amplitudes(vector, nkpts, (nmoa, nmob), (nocca, noccb))
    Hr1a, Hr1b = Hr1
    Hr2aaa, Hr2aba, Hr2bab, Hr2bbb = Hr2

    print('ea Hr1a',  abs(lib.fp(Hr1a)   - (-0.081373075311041126-0.51422895644026023j)))
    print('ea Hr1b',  abs(lib.fp(Hr1b)   - (-0.39518588661294807-1.3063424820239824j))  )
    print('ea Hr2aaa',abs(lib.fp(Hr2aaa) - (-2.6502079691200251-0.61302655915003545j))  )
    print('ea Hr2aba',abs(lib.fp(Hr2aba) - (5.5723208649566036-5.4202659143496286j))    )
    print('ea Hr2bab',abs(lib.fp(Hr2bab) - (-1.2822293707887937+0.3026476580141586j))   )
    print('ea Hr2bbb',abs(lib.fp(Hr2bbb) - (-4.0202809577487253-0.46985725132191702j))  )



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Xing Zhang <zhangxing.nju@gmail.com>
#

import numpy as np
import ctypes
from pyscf import lib
from pyscf.lib import logger
from pyscf import __config__
from pyscf.pbc.symm import symmetry as symm
from pyscf.pbc.lib.kpts_helper import member, KPT_DIFF_TOL, KptsHelper
from numpy.linalg import inv

libpbc = lib.load_library('libpbc')

def make_kpts_ibz(kpts):
    """
    Locate k-points in IBZ.

    Parameters
    ----------
        kpts : :class:`KPoints` object

    Notes
    -----
        This function modifies the :obj:`kpts` object.
    """
    cell = kpts.cell
    nkpts = kpts.nkpts
    nop = kpts.nop
    op_rot = np.asarray([op.a2b(cell).rot for op in kpts.ops])
    if kpts.time_reversal:
        op_rot = np.concatenate([op_rot, -op_rot])

    bz2bz_ks = map_k_points_fast(kpts.kpts_scaled, op_rot, KPT_DIFF_TOL)
    kpts.k2opk = bz2bz_ks.copy()
    if -1 in bz2bz_ks:
        bz2bz_ks[:, np.unique(np.where(bz2bz_ks == -1)[1])] = -1
        if kpts.verbose >= logger.WARN:
            logger.warn(kpts, 'k-points have lower symmetry than lattice.')

    bz2bz_k = -np.ones(nkpts+1, dtype=int)
    ibz2bz_k = []
    for k in range(nkpts-1, -1, -1):
        if bz2bz_k[k] == -1:
            bz2bz_k[bz2bz_ks[k]] = k
            ibz2bz_k.append(k)
    ibz2bz_k = np.array(ibz2bz_k[::-1])
    bz2bz_k = bz2bz_k[:-1].copy()

    bz2ibz_k = np.empty(nkpts, int)
    bz2ibz_k[ibz2bz_k] = np.arange(len(ibz2bz_k))
    bz2ibz_k = bz2ibz_k[bz2bz_k]

    kpts.bz2ibz = bz2ibz_k
    kpts.ibz2bz = ibz2bz_k
    kpts.weights_ibz = np.bincount(bz2ibz_k) * (1.0 / nkpts)
    kpts.kpts_scaled_ibz = kpts.kpts_scaled[kpts.ibz2bz]
    kpts.kpts_ibz = kpts.cell.get_abs_kpts(kpts.kpts_scaled_ibz)
    kpts.nkpts_ibz = len(kpts.kpts_ibz)

    for k in range(nkpts):
        bz_k_scaled = kpts.kpts_scaled[k]
        ibz_idx = kpts.bz2ibz[k]
        ibz_k_scaled = kpts.kpts_scaled_ibz[ibz_idx]
        for io, op in enumerate(op_rot):
            if -1 in bz2bz_ks[:,io]:
                #This rotation is not in the subgroup
                #that the k-mesh belongs to; only happens
                #when k-mesh has lower symmetry than lattice.
                continue
            diff = bz_k_scaled - np.dot(ibz_k_scaled, op.T)
            diff = diff - diff.round()
            if (np.absolute(diff) < KPT_DIFF_TOL).all():
                kpts.time_reversal_symm_bz[k] = io // nop
                kpts.stars_ops_bz[k] = io % nop
                break

    for i in range(kpts.nkpts_ibz):
        ibz_k_scaled = kpts.kpts_scaled_ibz[i]
        idx = np.where(kpts.bz2ibz == i)[0]
        kpts.stars.append(idx)
        kpts.stars_ops.append(kpts.stars_ops_bz[idx])

def make_ktuples_ibz(kpts, kpts_scaled=None, ntuple=2, tol=KPT_DIFF_TOL):
    """
    Constructe k-point tuples in IBZ.

    Parameters
    ----------
        kpts : :class:`KPoints` object
        kpts_scaled : (nkpts, ntuple, 3) ndarray
            Input k-points among which symmetry relations are seeked.
            Default is None, meaning all the k-points in :obj:`kpts` are considered.
        ntuple : int
             Dimension of the tuples. Default is 2.
        tol : float
            K-points differ by ``tol`` are considered as different.
            Default is 1e-6.

    Returns
    -------
        ibz2bz_kk : (nibz,) ndarray of int
            Mapping table from IBZ to full BZ.
        ibz_kk_weight : (nibz,) ndarray of int
            Weights of each k-point tuple in the IBZ.
        bz2ibz_kk : (nkpts**ntuple,) ndarray of int
            Mapping table from full BZ to IBZ.
        kk_group : list of (nk,) ndarrays of int
            Similar as :attr:`.stars`.
        kk_sym_group : list of (nk,) ndarrays of int
            Similar as :attr:`.stars_ops`.
    """
    if kpts_scaled is not None:
        cell = kpts.cell
        op_rot = np.asarray([op.a2b(cell).rot for op in kpts.ops])
        if kpts.time_reversal:
            op_rot = np.concatenate([op_rot, -op_rot])
        bz2bz_ksks = map_k_tuples(kpts_scaled, op_rot, ntuple=ntuple, tol=tol)
        bz2bz_ksks[:, np.unique(np.where(bz2bz_ksks == -1)[1])] = -1
        nbzk2 = bz2bz_ksks.shape[0]
    else:
        bz2bz_ks = kpts.k2opk
        nop = bz2bz_ks.shape[-1]
        nbzk = kpts.nkpts
        nbzk2 = nbzk**ntuple
        bz2bz_T = np.zeros([nop, nbzk2], dtype=int)
        for iop in range(nop):
            tmp = lib.cartesian_prod([bz2bz_ks[:,iop]]*ntuple)
            if -1 in tmp:
                bz2bz_T[iop,:] = -1
            else:
                #idx_throw = np.unique(np.where(tmp == -1)[0])
                for i in range(ntuple):
                    bz2bz_T[iop] += tmp[:,i] * nbzk**(ntuple-i-1)
                #bz2bz_T[iop, idx_throw] = -1
        bz2bz_ksks = bz2bz_T.T

    if kpts.verbose >= logger.INFO:
        logger.info(kpts, 'Number of k-point %d-tuples: %d', ntuple, nbzk2)

    bz2bz_kk = -np.ones(nbzk2+1, dtype=int)
    ibz2bz_kk = []
    k_group = []
    sym_group = []
    for k in range(nbzk2-1, -1, -1):
        if bz2bz_kk[k] == -1:
            bz2bz_kk[bz2bz_ksks[k]] = k
            ibz2bz_kk.append(k)
            k_idx, op_idx = np.unique(bz2bz_ksks[k], return_index=True)
            if k_idx[0] == -1:
                k_idx = k_idx[1:]
                op_idx = op_idx[1:]
            k_group.append(k_idx)
            sym_group.append(op_idx)

    ibz2bz_kk = np.array(ibz2bz_kk[::-1])
    if kpts.verbose >= logger.INFO:
        logger.info(kpts, 'Number of k %d-tuples in IBZ: %s', ntuple, len(ibz2bz_kk))

    bz2bz_kk = bz2bz_kk[:-1].copy()
    bz2ibz_kk = np.empty(nbzk2,dtype=int)
    bz2ibz_kk[ibz2bz_kk] = np.arange(len(ibz2bz_kk))
    bz2ibz_kk = bz2ibz_kk[bz2bz_kk]

    kk_group = k_group[::-1]
    kk_sym_group = sym_group[::-1]
    ibz_kk_weight = np.bincount(bz2ibz_kk) * (1.0 / nbzk2)
    return ibz2bz_kk, ibz_kk_weight, bz2ibz_kk, kk_group, kk_sym_group

def make_k4_ibz(kpts, sym='s1'):
    #physicist's notation
    ibz2bz, weight, bz2ibz, group, _ = kpts.make_ktuples_ibz(ntuple=3)
    khelper = KptsHelper(kpts.cell, kpts.kpts)
    k4 = []
    for ki, kj, ka in kpts.loop_ktuples(ibz2bz, 3):
        kb = khelper.kconserv[ki,ka,kj]
        k4.append([ki,kj,ka,kb])

    if sym == "s1":
        return np.asarray(k4), np.asarray(weight), np.asarray(bz2ibz)
    elif sym == "s2" or sym == "s4":
        ibz2ibz_s2 = np.arange(len(k4))
        k4_s2 = []
        weight_s2 = []
        for i, k in enumerate(k4):
            ki,kj,ka,kb = k
            k_sym = [kj,ki,kb,ka] #interchange dummy indices
            if k not in k4_s2 and k_sym not in k4_s2:
                k4_s2.append(k)
                ibz2ibz_s2[i] = len(k4_s2) - 1
                w = weight[i]
                if k != k_sym and k_sym in k4:
                    idx = k4.index(k_sym)
                    ibz2ibz_s2[idx] = ibz2ibz_s2[i]
                    w += weight[idx]
                weight_s2.append(w)
        #refine s2 symmetry
        k4_s2_refine = []
        weight_s2_refine = []
        skip = np.zeros([len(k4_s2)], dtype=int)
        ibz_s22ibz_s2_refine = np.arange(len(k4_s2))
        for i, k in enumerate(k4_s2):
            if skip[i]: continue
            ki,kj,ka,kb = k
            k_sym = [kj,ki,kb,ka]
            if ki==kj and ka==kb:
                k4_s2_refine.append(k)
                ibz_s22ibz_s2_refine[i] = len(k4_s2_refine) - 1
                weight_s2_refine.append(weight_s2[i])
                continue
            idx_sym = None
            for j in range(i+1, len(k4_s2)):
                if skip[j]: continue
                k_tmp = k4_s2[j]
                if ki in k_tmp and kj in k_tmp and ka in k_tmp and kb in k_tmp:
                    idx = k4.index(k_tmp)
                    for kii,kjj,kaa in kpts.loop_ktuples(group[idx], 3):
                        kbb = khelper.kconserv[kii,kaa,kjj]
                        if k_sym == [kii,kjj,kaa,kbb]:
                            idx_sym = j
                            break
                    if idx_sym is not None:
                        break
            w = weight_s2[i]
            if idx_sym is not None:
                skip[idx_sym] = 1
                w += weight_s2[idx_sym]
                ibz_s22ibz_s2_refine[idx_sym] = len(k4_s2_refine)
            k4_s2_refine.append(k)
            ibz_s22ibz_s2_refine[i] = len(k4_s2_refine) - 1
            weight_s2_refine.append(w)
        k4_s2 = k4_s2_refine
        weight_s2 = weight_s2_refine
        #end refine
        if sym == "s2":
            k4_s2 = np.asarray(k4_s2)
            weight_s2 = np.asarray(weight_s2)
            idx = np.lexsort(k4_s2.T[::-1,:])
            bz2ibz_s2 = np.arange(len(bz2ibz))
            for i in range(len(bz2ibz)):
                bz2ibz_s2[i] = np.where(idx == ibz_s22ibz_s2_refine[ibz2ibz_s2[bz2ibz[i]]])[0]
            return k4_s2[idx], weight_s2[idx], bz2ibz_s2
        else:
            k4_s4 = []
            weight_s4 = []
            for i, k in enumerate(k4_s2):
                ki,kj,ka,kb = k
                k_sym = [ka,kb,ki,kj] #complex conjugate
                if k not in k4_s4 and k_sym not in k4_s4:
                    k4_s4.append(k)
                    w = weight_s2[i]
                    if k != k_sym and k_sym in k4_s2:
                        idx = k4_s2.index(k_sym)
                        w += weight_s2[idx]
                    weight_s4.append(w)
            k4_s4 = np.asarray(k4_s4)
            weight_s4 = np.asarray(weight_s4)
            idx = np.lexsort(k4_s4.T[::-1,:])
            return k4_s4[idx], weight_s4[idx], None
    else:
        raise NotImplementedError("Unsupported symmetry.")
    return

def map_k_points_fast(kpts_scaled, ops, tol=KPT_DIFF_TOL):
    #This routine is modified from GPAW
    """
    Find symmetry-related k-points.

    Parameters
    ----------
        kpts_scaled : (nkpts, 3) ndarray
            Scaled k-points.
        ops : (nop, 3, 3) ndarray of int
            Rotation operators.
        tol : float
            K-points differ by ``tol`` are considered as different.
            Default is 1e-6.

    Returns
    -------
        bz2bz_ks : (nkpts, nop) ndarray of int
            mapping table between k and op*k.
            bz2bz_ks[k1,s] = k2 if ops[s] * kpts_scaled[k1] = kpts_scaled[k2] + K,
            where K is a reciprocal lattice vector.
    """
    nkpts = len(kpts_scaled)
    nop = len(ops)
    bz2bz_ks = -np.ones((nkpts, nop), dtype=int)
    for s, op in enumerate(ops):
        # Find mapped kpoints
        op_kpts_scaled = np.dot(kpts_scaled, op.T)

        # Do some work on the input
        k_kc = np.concatenate([kpts_scaled, op_kpts_scaled])
        k_kc = np.mod(np.mod(k_kc, 1), 1)
        k_kc = aglomerate_points(k_kc, tol)
        k_kc = k_kc.round(-np.log10(tol).astype(int))
        k_kc = np.mod(k_kc, 1)

        # Find the lexicographical order
        order = np.lexsort(k_kc.T)
        k_kc = k_kc[order]
        diff_kc = np.diff(k_kc, axis=0)
        equivalentpairs_k = np.array((diff_kc == 0).all(1), dtype=bool)

        # Mapping array.
        orders = np.array([order[:-1][equivalentpairs_k],
                           order[1:][equivalentpairs_k]])

        # This has to be true.
        assert (orders[0] < nkpts).all()
        assert (orders[1] >= nkpts).all()
        bz2bz_ks[orders[1] - nkpts, s] = orders[0]
    return bz2bz_ks

def map_k_tuples(kpts_scaled, ops, ntuple=2, tol=KPT_DIFF_TOL):
    """
    Find symmetry-related k-point tuples.

    Parameters
    ----------
        kpts_scaled : (nkpts, ntuple, 3) ndarray
            Scaled k-point tuples.
        ops : (nop, 3, 3) ndarray of int
            Rotation operators.
        ntuple : int
            Dimension of tuples. Default is 2.
        tol : float
            K-points differ by ``tol`` are considered as different.
            Default is 1e-6.

    Returns
    -------
        bz2bz_ks : (nkpts, nop) ndarray of int
            mapping table between k and op*k.
            bz2bz_ks[k1,s] = k2 if ops[s] * kpts_scaled[k1] = kpts_scaled[k2] + K,
            where K is a reciprocal lattice vector.
    """
    nkpts = len(kpts_scaled)
    nop = len(ops)
    bz2bz_ks = -np.ones((nkpts, nop), dtype=int)
    for s, op in enumerate(ops):
        # Find mapped kpoints
        op_kpts_scaled = np.empty((nkpts, ntuple, 3), dtype=float)
        for i in range(ntuple):
            op_kpts_scaled[:,i] = np.dot(kpts_scaled[:,i], op.T)
        op_kpts_scaled = op_kpts_scaled.reshape((nkpts,-1))

        # Do some work on the input
        k_kc = np.concatenate([kpts_scaled.reshape((nkpts,-1)), op_kpts_scaled])
        k_kc = np.mod(np.mod(k_kc, 1), 1)
        k_kc = aglomerate_points(k_kc, tol)
        k_kc = k_kc.round(-np.log10(tol).astype(int))
        k_kc = np.mod(k_kc, 1)

        # Find the lexicographical order
        order = np.lexsort(k_kc.T)
        k_kc = k_kc[order]
        diff_kc = np.diff(k_kc, axis=0)
        equivalentpairs_k = np.array((diff_kc == 0).all(1), dtype=bool)

        # Mapping array.
        orders = np.array([order[:-1][equivalentpairs_k],
                           order[1:][equivalentpairs_k]])

        # This has to be true.
        assert (orders[0] < nkpts).all()
        assert (orders[1] >= nkpts).all()
        bz2bz_ks[orders[1] - nkpts, s] = orders[0]
    return bz2bz_ks

def aglomerate_points(k_kc, tol=KPT_DIFF_TOL):
    #This routine is adopted from GPAW
    '''
    Remove numerical error
    '''
    nd = k_kc.shape[1]
    nbzkpts = len(k_kc)

    inds_kc = np.argsort(k_kc, axis=0)

    for c in range(nd):
        sk_k = k_kc[inds_kc[:, c], c]
        dk_k = np.diff(sk_k)

        pt_K = np.argwhere(dk_k > tol)[:, 0]
        pt_K = np.append(np.append(0, pt_K + 1), nbzkpts*2)
        for i in range(len(pt_K) - 1):
            k_kc[inds_kc[pt_K[i]:pt_K[i + 1], c], c] = k_kc[inds_kc[pt_K[i], c], c]
    return k_kc

def symmetrize_density(kpts, rhoR_k, ibz_k_idx, mesh):
    '''
    Transform real-space densities from IBZ to full BZ
    '''
    rhoR_k = np.asarray(rhoR_k, order='C')
    rhoR = np.zeros_like(rhoR_k, order='C')

    dtype = rhoR_k.dtype
    if dtype == np.double:
        symmetrize = libpbc.symmetrize
        symmetrize_ft = libpbc.symmetrize_ft
    elif dtype == np.complex128:
        symmetrize = libpbc.symmetrize_complex
        symmetrize_ft = libpbc.symmetrize_ft_complex
    else:
        raise RuntimeError("Unsupported data type %s" % dtype)

    c_rhoR = rhoR.ctypes.data_as(ctypes.c_void_p)
    c_rhoR_k = rhoR_k.ctypes.data_as(ctypes.c_void_p)

    mesh = np.asarray(mesh, dtype=np.int32, order='C')
    c_mesh = mesh.ctypes.data_as(ctypes.c_void_p)
    for iop in kpts.stars_ops[ibz_k_idx]:
        op = kpts.ops[iop]
        if op.is_eye: #or op.is_inversion:
            rhoR += rhoR_k
        else:
            inv_op = op.inv()
            op_rot = np.asarray(inv_op.rot, dtype=np.int32, order='C')
            c_op_rot = op_rot.ctypes.data_as(ctypes.c_void_p)
            if inv_op.trans_is_zero:
                symmetrize(c_rhoR, c_rhoR_k, c_op_rot, c_mesh)
            else:
                trans = np.asarray(inv_op.trans, dtype=np.double, order='C')
                c_trans = trans.ctypes.data_as(ctypes.c_void_p)
                symmetrize_ft(c_rhoR, c_rhoR_k, c_op_rot, c_trans, c_mesh)
    return rhoR

def symmetrize_wavefunction(kpts, psiR_k, mesh):
    '''
    transform real-space wavefunctions from IBZ to full BZ
    '''
    raise RuntimeError('need verification')
    psiR_k = np.asarray(psiR_k, order='C')
    is_complex = psiR_k.dtype == np.complex
    nao = psiR_k.shape[1]
    nG = psiR_k.shape[2]
    psiR = np.zeros([kpts.nkpts,nao,nG], dtype = psiR_k.dtype, order='C')

    mesh = np.asarray(mesh, dtype=np.int32, order='C')
    c_mesh = mesh.ctypes.data_as(ctypes.c_void_p)

    for ibz_k_idx in range(kpts.nkpts_ibz):
        for idx, iop in enumerate(kpts.stars_ops[ibz_k_idx]):
            bz_k_idx = kpts.stars[ibz_k_idx][idx]
            op = kpts.ops[iop].b2a(kpts.cell).rot
            op = np.asarray(op, dtype=np.int32, order='C')
            time_reversal = kpts.time_reversal_symm_bz[bz_k_idx]
            if symm.is_eye(op): #or symm.is_inversion(op):
                psiR[bz_k_idx] = psiR_k[ibz_k_idx]
            else:
                c_psiR = psiR[bz_k_idx].ctypes.data_as(ctypes.c_void_p)
                c_psiR_k = psiR_k[ibz_k_idx].ctypes.data_as(ctypes.c_void_p)
                c_op = op.ctypes.data_as(ctypes.c_void_p)
                if is_complex:
                    libpbc.symmetrize_complex(c_psiR, c_psiR_k, c_op, c_mesh)
                else:
                    libpbc.symmetrize(c_psiR, c_psiR_k, c_op, c_mesh)
            if time_reversal and is_complex:
                psiR[bz_k_idx] = psiR[bz_k_idx].conj()
    return psiR

def transform_mo_coeff(kpts, mo_coeff_ibz):
    """
    Transform MO coefficients from IBZ to full BZ.

    Parameters
    ----------
        kpts : :class:`KPoints` object
        mo_coeff_ibz : ([2,] nkpts_ibz, nao, nmo) ndarray
            MO coefficients for k-points in IBZ.

    Returns
    -------
        mo_coeff_bz : ([2,] nkpts, nao, nmo) ndarray
            MO coefficients for k-points in full BZ.
    """
    mos = []
    is_uhf = False
    if isinstance(mo_coeff_ibz[0][0], np.ndarray) and mo_coeff_ibz[0][0].ndim == 2:
        is_uhf = True
        mos = [[],[]]
    for k in range(kpts.nkpts):
        ibz_k_idx = kpts.bz2ibz[k]
        ibz_k_scaled = kpts.kpts_scaled_ibz[ibz_k_idx]
        iop = kpts.stars_ops_bz[k]
        op = kpts.ops[iop]
        time_reversal = kpts.time_reversal_symm_bz[k]

        def _transform(mo_ibz, iop, op):
            if op.is_eye:
                mo_bz = mo_ibz
            else:
                mo_bz = symm.transform_mo_coeff(kpts.cell, ibz_k_scaled, mo_ibz, op, kpts.Dmats[iop])
            if time_reversal:
                mo_bz = mo_bz.conj()
            return mo_bz

        if is_uhf:
            mo_coeff_a = mo_coeff_ibz[0][ibz_k_idx]
            mos[0].append(_transform(mo_coeff_a, iop, op))
            mo_coeff_b = mo_coeff_ibz[1][ibz_k_idx]
            mos[1].append(_transform(mo_coeff_b, iop, op))
        else:
            mo_coeff = mo_coeff_ibz[ibz_k_idx]
            mos.append(_transform(mo_coeff, iop, op))
    return mos

def transform_mo_coeff_k(kpts, mo_coeff_ibz, k):
    """
    Get MO coefficients for a single k-point in full BZ.

    Parameters
    ----------
        kpts : :class:`KPoints` object
        mo_coeff_ibz : (nkpts_ibz, nao, nmo) ndarray
            MO coefficients for k-points in IBZ.
        k : int
            Index of the k-point in full BZ.

    Returns
    -------
        mo_coeff_bz : (nao, nmo) ndarray
            MO coefficients for the ``k``-th k-point in full BZ.
    """
    ibz_k_idx = kpts.bz2ibz[k]
    ibz_k_scaled = kpts.kpts_scaled_ibz[ibz_k_idx]
    iop = kpts.stars_ops_bz[k]
    op = kpts.ops[iop]
    time_reversal = kpts.time_reversal_symm_bz[k]

    mo_ibz = mo_coeff_ibz[ibz_k_idx]
    if op.is_eye:
        mo_bz = mo_ibz
    else:
        mo_bz = symm.transform_mo_coeff(kpts.cell, ibz_k_scaled, mo_ibz, op, kpts.Dmats[iop])
    if time_reversal:
        mo_bz = mo_bz.conj()
    return mo_bz

transform_single_mo_coeff = transform_mo_coeff_k

def transform_mo_occ(kpts, mo_occ_ibz):
    '''
    Transform MO occupations from IBZ to full BZ
    '''
    occ = []
    is_uhf = False
    if isinstance(mo_occ_ibz[0][0], np.ndarray) and mo_occ_ibz[0][0].ndim == 1:
        is_uhf = True
        occ = [[],[]]
    for k in range(kpts.nkpts):
        ibz_k_idx = kpts.bz2ibz[k]
        if is_uhf:
            occ[0].append(mo_occ_ibz[0][ibz_k_idx])
            occ[1].append(mo_occ_ibz[1][ibz_k_idx])
        else:
            occ.append(mo_occ_ibz[ibz_k_idx])
    return occ

def transform_dm(kpts, dm_ibz):
    """
    Transform density matrices from IBZ to full BZ.

    Parameters
    ----------
        kpts : :class:`KPoints` object
        dm_ibz : ([2,] nkpts_ibz, nao, nao) ndarray
            Density matrices for k-points in IBZ.

    Returns
    -------
        dm_bz : ([2,] nkpts, nao, nao) ndarray
            Density matrices for k-points in full BZ.
    """
    mo_occ = mo_coeff = None
    if getattr(dm_ibz, "mo_coeff", None) is not None:
        mo_coeff = kpts.transform_mo_coeff(dm_ibz.mo_coeff)
        mo_occ = kpts.transform_mo_occ(dm_ibz.mo_occ)

    dms = []
    is_uhf = False
    if (isinstance(dm_ibz, np.ndarray) and dm_ibz.ndim == 4) or \
       (isinstance(dm_ibz[0][0], np.ndarray) and dm_ibz[0][0].ndim == 2):
        is_uhf = True
        dms = [[],[]]
    for k in range(kpts.nkpts):
        ibz_k_idx = kpts.bz2ibz[k]
        ibz_kpt_scaled = kpts.kpts_scaled_ibz[ibz_k_idx]
        iop = kpts.stars_ops_bz[k]
        op = kpts.ops[iop]
        time_reversal = kpts.time_reversal_symm_bz[k]

        def _transform(dm_ibz, iop, op):
            if op.is_eye:
                dm_bz = dm_ibz
            else:
                dm_bz = symm.transform_dm(kpts.cell, ibz_kpt_scaled, dm_ibz, op, kpts.Dmats[iop])
            if time_reversal:
                dm_bz = dm_bz.conj()
            return dm_bz

        if is_uhf:
            dm_a = dm_ibz[0][ibz_k_idx]
            dms[0].append(_transform(dm_a, iop, op))
            dm_b = dm_ibz[1][ibz_k_idx]
            dms[1].append(_transform(dm_b, iop, op))
        else:
            dm = dm_ibz[ibz_k_idx]
            dms.append(_transform(dm, iop, op))
    if is_uhf:
        nkpts = len(dms[0])
        nao = dms[0][0].shape[0]
        dms = lib.asarray(dms).reshape(2,nkpts,nao,nao)
    else:
        dms = lib.asarray(dms)
    return lib.tag_array(dms, mo_coeff=mo_coeff, mo_occ=mo_occ)

def dm_at_ref_cell(kpts, dm_ibz):
    """
    Given the density matrices in IBZ, compute the reference cell density matrix.

    Parameters
    ----------
        kpts : :class:`KPoints` object
        dm_ibz : ([2,] nkpts_ibz, nao, nao) ndarray
            Density matrices for k-points in IBZ.

    Returns
    -------
        dm0 : ([2,] nao, nao) ndarray
            Density matrix at reference cell.
    """
    dm_bz = transform_dm(kpts, dm_ibz)
    dm0 = np.sum(dm_bz, axis=-3) / kpts.nkpts
    if abs(dm0.imag).max() > 1e-10:
        logger.warn(kpts, 'Imaginary density matrix found at reference cell: \
                    abs(dm0.imag).max() = %g', abs(dm0.imag).max())
    return dm0

def transform_mo_energy(kpts, mo_energy_ibz):
    '''
    Transform MO energies from IBZ to full BZ
    '''
    is_uhf = False
    if isinstance(mo_energy_ibz[0][0], np.ndarray):
        is_uhf = True
    mo_energy_bz = []
    if is_uhf:
        mo_energy_bz = [[],[]]
    for k in range(kpts.nkpts):
        ibz_k_idx = kpts.bz2ibz[k]
        if is_uhf:
            mo_energy_bz[0].append(mo_energy_ibz[0][ibz_k_idx])
            mo_energy_bz[1].append(mo_energy_ibz[1][ibz_k_idx])
        else:
            mo_energy_bz.append(mo_energy_ibz[ibz_k_idx])
    return mo_energy_bz

def transform_1e_operator(kpts, fock_ibz):
    """
    Transform 1-electron operator from IBZ to full BZ.

    Parameters
    ----------
        kpts : :class:`KPoints` object
        fock_ibz : ([2,] nkpts_ibz, nao, nao) ndarray
            Fock-like matrices for k-points in IBZ.

    Returns
    -------
        fock_bz : ([2,] nkpts, nao, nao) ndarray
            Fock-like matrices for k-points in full BZ.
    """
    fock = []
    is_uhf = False
    if isinstance(fock_ibz[0][0], np.ndarray) and fock_ibz[0][0].ndim == 2:
        is_uhf = True
        fock = [[],[]]

    for k in range(kpts.nkpts):
        ibz_k_idx = kpts.bz2ibz[k]
        ibz_kpt_scaled = kpts.kpts_scaled_ibz[ibz_k_idx]
        iop = kpts.stars_ops_bz[k]
        op = kpts.ops[iop]
        time_reversal = kpts.time_reversal_symm_bz[k]

        def _transform(fock_ibz, iop, op):
            if op.is_eye:
                fock_bz = fock_ibz
            else:
                fock_bz = symm.transform_1e_operator(kpts.cell, ibz_kpt_scaled, fock_ibz, op, kpts.Dmats[iop])
            if time_reversal:
                fock_bz = fock_bz.conj()
            return fock_bz

        if is_uhf:
            fock_a = fock_ibz[0][ibz_k_idx]
            fock[0].append(_transform(fock_a, iop, op))
            fock_b = fock_ibz[1][ibz_k_idx]
            fock[1].append(_transform(fock_b, iop, op))
        else:
            fock.append(_transform(fock_ibz[ibz_k_idx], iop, op))
    if is_uhf:
        nkpts = len(fock[0])
        nao = fock[0][0].shape[0]
        fock = lib.asarray(fock).reshape(2,nkpts,nao,nao)
    else:
        fock = lib.asarray(fock)
    return fock

transform_fock = transform_1e_operator

def check_mo_occ_symmetry(kpts, mo_occ, tol=1e-5):
    """
    Check if MO occupations in full BZ have the correct symmetry.
    If not, raise error; else, return MO occupations in IBZ.

    Parameters
    ----------
        kpts : :class:`KPoints` object
        mo_occ : list of (nmo,) ndarray
            MO occupations for k-points in full BZ.
            len(mo_occ) = nkpts
        tol : float
            Occupations differ less than ``tol`` are considered as the same.
            Default is 1e-5.

    Returns
    -------
        mo_occ_ibz : list of (nmo,) ndarray
            MO occupations for k-points in IBZ.
            len(mo_occ_ibz) = nkpts_ibz

    Raises
    ------
        RuntimeError
            If symmetry is broken.
    """
    for bz_k in kpts.stars:
        nbzk = len(bz_k)
        for i in range(nbzk):
            for j in range(i+1,nbzk):
                if not (np.absolute(mo_occ[bz_k[i]] - mo_occ[bz_k[j]]) < tol).all():
                    raise RuntimeError("Symmetry broken solution found. \
                                        This is probably due to KUHF calculations \
                                        with integer occupation numbers. \
                                        Try use smearing or turn off symmetry.")
    mo_occ_ibz = []
    for k in range(kpts.nkpts_ibz):
        mo_occ_ibz.append(mo_occ[kpts.ibz2bz[k]])
    return mo_occ_ibz

def make_kpts(cell, kpts=np.zeros((1,3)),
              space_group_symmetry=False, time_reversal_symmetry=False,
              symmorphic=True):
    """
    A wrapper function to build the :class:`KPoints` object.

    Parameters
    ----------
        cell : :class:`Cell` instance
            Unit cell information.
        kpts : (nkpts,3) ndarray
            K-points in full BZ.
        space_group_symmetry : bool
            Whether to consider space group symmetry. Default is False.
        time_reversal_symmetry : bool
            Whether to consider time reversal symmetry. Default is False.
        symmorphic : bool
            Whether to consider only the symmorphic subgroup. Default is True.

    Examples
    --------
    >>> cell = gto.M(
    ...     atom = '''He 0. 0. 0.''',
    ...     a = numpy.eye(3)*2.0).build()
    >>> kpts = make_kpts(cell,
    ...                  numpy.array([[0.,0.,0.],[0.5,0.,0.],[0.,0.5,0.],[0.,0.,0.5]])
    ...                  space_group_symmetry=True)
    >>> print(kpts.kpts_ibz)
    [[0.  0.  0. ]
     [0.  0.  0.5]]
    """
    if isinstance(kpts, KPoints):
        return kpts.build(space_group_symmetry, time_reversal_symmetry, symmorphic)
    else:
        return KPoints(cell, kpts).build(space_group_symmetry, time_reversal_symmetry, symmorphic)

class KPoints(symm.Symmetry, lib.StreamObject):
    """
    A symmetry object which handles k-point symmetry.

    Parameters
    ----------
        cell : :class:`Cell` instance
            Unit cell information.
        kpts : (nkpts,3) ndarray
            K-points in full BZ.

    Examples
    --------
    >>> cell = gto.M(
    ...     atom = '''He 0. 0. 0.''',
    ...     a = numpy.eye(3)*2.0).build()
    >>> kpts = KPoints(cell, numpy.array([[0.,0.,0.],[0.5,0.,0.],[0.,0.5,0.],[0.,0.,0.5]]))
    >>> kpts.build(space_group_symmetry=True)
    >>> print(kpts.kpts_ibz)
    [[0.  0.  0. ]
     [0.  0.  0.5]]

    Attributes
    ----------
        cell : :class:`Cell` instance
            Unit cell information.
        verbose : int
            Print level. Default value is ``cell.verbose``.
        time_reversal : bool
            Whether to consider time-reversal symmetry.
            For systems with inversion symmetry, time-reversal symmetry is not considered
            unless spin-orbit coupling is present.
        kpts : (nkpts,3) ndarray
            K-points in full BZ.
        kpts_scaled : (nkpts,3) ndarray
            Scaled k-points in full BZ.
        weights : (nkpts,) ndarray
            Weights of k-points in full BZ.
        bz2ibz : (nkpts,) ndarray of int
            Mapping table from full BZ to IBZ.
        kpts_ibz : (nkpts_ibz,3) ndarray
            K-points in IBZ.
        kpts_scaled_ibz : (nkpts_ibz,3) ndarray
            Scaled k-points in IBZ.
        weights_ibz : (nkpts_ibz,) ndarray
            Weights of k-points in IBZ.
        ibz2bz : (nkpts_ibz,) ndarray of int
            Mapping table from IBZ to full BZ.
        k2opk (bz2bz_ks) : (nkpts, nop*(time_reversal+1)) ndarray of int
            Mapping table between kpts and ops.rot * kpts.
        stars : list of (nk,) ndarrays of int
            Stars of k-points in full BZ with len(stars) = nkpts_ibz
            and ``nk`` is the number of symmetry-related k-points for each k-point in IBZ.
        stars_ops : same as ``stars``
            Indices of rotation operators connecting k points in full BZ and in IBZ.
        stars_ops_bz : (nkpts,) ndarray of int
            Same as stars_ops but arranged in the sequence of k-points in full BZ.
        time_reversal_symm_bz : (nkpts,) ndarray of int
            Whether k-points in BZ and IBZ are related in addition by time-reversal symmetry.
    """
    def __init__(self, cell=None, kpts=np.zeros((1,3))):
        symm.Symmetry.__init__(self, cell)
        self.verbose = getattr(cell, 'verbose', logger.NOTE)
        self.stdout = getattr(cell, 'stdout', None)
        self.time_reversal = False

        self.kpts_ibz = self.kpts = kpts
        self.kpts_scaled_ibz = self.kpts_scaled = None
        nkpts = len(self.kpts)
        self.weights_ibz = self.weights = np.asarray([1./nkpts] * nkpts)
        self.ibz2bz = self.bz2ibz = np.arange(nkpts, dtype=int)

        self.k2opk = None
        self.stars = []
        self.stars_ops = []
        self.stars_ops_bz = np.zeros(nkpts, dtype=int)
        self.time_reversal_symm_bz = np.zeros(nkpts, dtype=int)

        #private variables
        self._nkpts = len(self.kpts)
        self._nkpts_ibz = len(self.kpts_ibz)

    @property
    def nkpts(self):
        return self._nkpts

    @nkpts.setter
    def nkpts(self, n):
        self._nkpts = n

    @property
    def nkpts_ibz(self):
        return self._nkpts_ibz

    @nkpts_ibz.setter
    def nkpts_ibz(self, n):
        self._nkpts_ibz = n

    def __str__(self):
        s = ''
        s += self.__repr__()

        s += '\nk-points (scaled) in BZ                        weights\n'
        bzk = self.kpts_scaled
        for k in range(self.nkpts):
            s += '%d:  %11.8f, %11.8f, %11.8f    %9.6f\n' % \
                (k, bzk[k][0], bzk[k][1], bzk[k][2], self.weights[k])

        s += 'k-points (scaled) in IBZ                       weights\n'
        ibzk = self.kpts_scaled_ibz
        for k in range(self.nkpts_ibz):
            s += '%d:  %11.8f, %11.8f, %11.8f    %9.6f\n' % \
                (k, ibzk[k][0], ibzk[k][1], ibzk[k][2], self.weights_ibz[k])

        s += 'mapping from BZ to IBZ\n'
        s += "%s" % self.bz2ibz
        s += '\nmapping from IBZ to BZ\n'
        s += "%s" % self.ibz2bz
        return s

    def __len__(self):
        return self.nkpts_ibz

    def build(self, space_group_symmetry=True, time_reversal_symmetry=True,
              symmorphic=True, *args, **kwargs):
        symm.Symmetry.build(self, space_group_symmetry, symmorphic, *args, **kwargs)
        if not getattr(self.cell, '_built', None): return

        self.time_reversal = time_reversal_symmetry and not self.has_inversion
        self.kpts_scaled_ibz = self.kpts_scaled = self.cell.get_scaled_kpts(self.kpts)
        self.make_kpts_ibz()
        self.dump_info()
        return self

    def dump_info(self):
        if self.verbose >= logger.INFO:
            logger.info(self, 'time reversal: %s', self.time_reversal)
            logger.info(self, 'k-points in IBZ                           weights')
            ibzk = self.kpts_scaled_ibz
            for k in range(self.nkpts_ibz):
                logger.info(self, '%d:  %11.8f, %11.8f, %11.8f    %d/%d',
                            k, ibzk[k][0], ibzk[k][1], ibzk[k][2],
                            np.floor(self.weights_ibz[k]*self.nkpts), self.nkpts)

    def make_gdf_kptij_lst_jk(self):
        '''
        Build GDF k-point-pair list for get_jk
        All combinations:
            k_ibz != k_bz
            k_bz  == k_bz
        '''
        kptij_lst = [(self.kpts[i], self.kpts[i]) for i in range(self.nkpts)]
        for i in range(self.nkpts_ibz):
            ki = self.kpts_ibz[i]
            where = member(ki, self.kpts)
            for j in range(self.nkpts):
                kj = self.kpts[j]
                if j not in where:
                    kptij_lst.extend([(ki,kj)])
        kptij_lst = np.asarray(kptij_lst)
        return kptij_lst

    def loop_ktuples(self, ibz2bz, ntuple):
        nkpts = self.nkpts
        for k in ibz2bz:
            res = []
            for i in range(ntuple-1, -1, -1):
                ki = k // nkpts**(i)
                k = k - ki * nkpts**(i)
                res.append(ki)
            yield tuple(res)

    make_kpts_ibz = make_kpts_ibz
    make_ktuples_ibz = make_ktuples_ibz
    make_k4_ibz = make_k4_ibz
    loop_ktuples = loop_ktuples
    symmetrize_density = symmetrize_density
    symmetrize_wavefunction = symmetrize_wavefunction
    transform_mo_coeff = transform_mo_coeff
    transform_single_mo_coeff = transform_single_mo_coeff
    transform_dm = transform_dm
    transform_mo_energy = transform_mo_energy
    transform_mo_occ = transform_mo_occ
    check_mo_occ_symmetry = check_mo_occ_symmetry
    transform_fock = transform_fock
    transform_1e_operator = transform_1e_operator
    dm_at_ref_cell = dm_at_ref_cell

if __name__ == "__main__":
    import numpy
    from pyscf.pbc import gto
    cell = gto.Cell()
    cell.atom = """
        Si  0.0 0.0 0.0
        Si  1.3467560987 1.3467560987 1.3467560987
    """
    cell.a = [[0.0, 2.6935121974, 2.6935121974],
              [2.6935121974, 0.0, 2.6935121974],
              [2.6935121974, 2.6935121974, 0.0]]
    cell.verbose = 4
    cell.build()
    nk = [3,3,3]
    kpts_bz = cell.make_kpts(nk)
    kpts0 = cell.make_kpts(nk, space_group_symmetry=True, time_reversal_symmetry=True)
    kpts1 = KPoints(cell, kpts_bz).build(space_group_symmetry=True, time_reversal_symmetry=True)
    print(numpy.allclose(kpts0.kpts_ibz, kpts1.kpts_ibz))

    kpts = KPoints()
    print(kpts.kpts)



"""
This module implements the G0W0 approximation on top of `pyscf.tdscf.rhf_slow` TDHF implementation. Unlike `gw.py`, all
integrals are stored in memory. Several variants of GW are available:

 * `pyscf.gw_slow`: the molecular implementation;
 * (this module) `pyscf.pbc.gw.gw_slow`: single-kpoint PBC (periodic boundary condition) implementation;
 * `pyscf.pbc.gw.kgw_slow_supercell`: a supercell approach to PBC implementation with multiple k-points. Runs the
   molecular code for a model with several k-points for the cost of discarding momentum conservation and using dense
   instead of sparse matrixes;
 * `pyscf.pbc.gw.kgw_slow`: a PBC implementation with multiple k-points;
"""

# Convention for these modules:
# * IMDS contains routines for intermediates
# * kernel finds GW roots
# * GW provides a container

# This module and kgw_slow_supercell are simply aliases of the molecular code
from pyscf.gw import gw_slow


IMDS = gw_slow.IMDS
kernel = gw_slow.kernel
GW = gw_slow.GW



#!/usr/bin/env python
# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Generalized Kohn-Sham
'''


import numpy
import scipy.linalg
from pyscf import lib
from pyscf.lib import logger
from pyscf.scf import ghf
from pyscf.dft import rks
from pyscf.dft.numint2c import NumInt2C


def get_veff(ks, mol=None, dm=None, dm_last=0, vhf_last=0, hermi=1):
    '''Coulomb + XC functional

    .. note::
        This function will change the ks object.

    Args:
        ks : an instance of :class:`RKS`
            XC functional are controlled by ks.xc attribute.  Attribute
            ks.grids might be initialized.
        dm : ndarray or list of ndarrays
            A density matrix or a list of density matrices

    Kwargs:
        dm_last : ndarray or a list of ndarrays or 0
            The density matrix baseline.  If not 0, this function computes the
            increment of HF potential w.r.t. the reference HF potential matrix.
        vhf_last : ndarray or a list of ndarrays or 0
            The reference Vxc potential matrix.
        hermi : int
            Whether J, K matrix is hermitian

            | 0 : no hermitian or symmetric
            | 1 : hermitian
            | 2 : anti-hermitian

    Returns:
        matrix Veff = J + Vxc.  Veff can be a list matrices, if the input
        dm is a list of density matrices.
    '''
    if mol is None: mol = ks.mol
    if dm is None: dm = ks.make_rdm1()
    ks.initialize_grids(mol, dm)

    t0 = (logger.process_clock(), logger.perf_counter())

    ground_state = isinstance(dm, numpy.ndarray) and dm.ndim == 2

    if hermi == 2:  # because rho = 0
        n, exc, vxc = 0, 0, 0
    else:
        max_memory = ks.max_memory - lib.current_memory()[0]
        ni = ks._numint
        n, exc, vxc = ni.get_vxc(mol, ks.grids, ks.xc, dm,
                                 hermi=hermi, max_memory=max_memory)
        if ks.nlc != '':
            assert ('VV10' in ks.nlc.upper())
            _, enlc, vnlc = ni.get_vxc(mol, ks.nlcgrids, ks.xc+'__'+ks.nlc, dm,
                                       hermi=hermi, max_memory=max_memory)
            exc += enlc
            vxc += vnlc
        logger.debug(ks, 'nelec by numeric integration = %s', n)
        t0 = logger.timer(ks, 'vxc', *t0)

    #enabling range-separated hybrids
    omega, alpha, hyb = ni.rsh_and_hybrid_coeff(ks.xc, spin=mol.spin)

    if abs(hyb) < 1e-10 and abs(alpha) < 1e-10:
        vk = None
        if (ks._eri is None and ks.direct_scf and
            getattr(vhf_last, 'vj', None) is not None):
            ddm = numpy.asarray(dm) - numpy.asarray(dm_last)
            vj = ks.get_j(mol, ddm, hermi)
            vj += vhf_last.vj
        else:
            vj = ks.get_j(mol, dm, hermi)
        vxc += vj
    else:
        if (ks._eri is None and ks.direct_scf and
            getattr(vhf_last, 'vk', None) is not None):
            ddm = numpy.asarray(dm) - numpy.asarray(dm_last)
            vj, vk = ks.get_jk(mol, ddm, hermi)
            vk *= hyb
            if abs(omega) > 1e-10:
                vklr = ks.get_k(mol, ddm, hermi, omega=omega)
                vklr *= (alpha - hyb)
                vk += vklr
            vj += vhf_last.vj
            vk += vhf_last.vk
        else:
            vj, vk = ks.get_jk(mol, dm, hermi)
            vk *= hyb
            if abs(omega) > 1e-10:
                vklr = ks.get_k(mol, dm, hermi, omega=omega)
                vklr *= (alpha - hyb)
                vk += vklr
        vxc += vj - vk

        if ground_state:
            exc -= numpy.einsum('ij,ji', dm, vk).real * .5
    if ground_state:
        ecoul = numpy.einsum('ij,ji', dm, vj).real * .5
    else:
        ecoul = None

    vxc = lib.tag_array(vxc, ecoul=ecoul, exc=exc, vj=vj, vk=vk)
    return vxc

energy_elec = rks.energy_elec


class GKS(rks.KohnShamDFT, ghf.GHF):
    '''Generalized Kohn-Sham'''
    def __init__(self, mol, xc='LDA,VWN'):
        ghf.GHF.__init__(self, mol)
        rks.KohnShamDFT.__init__(self, xc)
        self._numint = NumInt2C()

    def dump_flags(self, verbose=None):
        ghf.GHF.dump_flags(self, verbose)
        rks.KohnShamDFT.dump_flags(self, verbose)
        logger.info(self, 'collinear = %s', self._numint.collinear)
        if self._numint.collinear[0] == 'm':
            logger.info(self, 'mcfun spin_samples = %s', self._numint.spin_samples)
            logger.info(self, 'mcfun collinear_thrd = %s', self._numint.collinear_thrd)
            logger.info(self, 'mcfun collinear_samples = %s', self._numint.collinear_samples)
        return self

    get_veff = get_veff
    energy_elec = energy_elec

    @property
    def collinear(self):
        return self._numint.collinear
    @collinear.setter
    def collinear(self, val):
        self._numint.collinear = val

    @property
    def spin_samples(self):
        return self._numint.spin_samples
    @spin_samples.setter
    def spin_samples(self, val):
        self._numint.spin_samples = val

    def nuc_grad_method(self):
        raise NotImplementedError


if __name__ == '__main__':
    from pyscf import gto
    mol = gto.Mole()
    mol.verbose = 3
    mol.atom = 'H 0 0 0; H 0 0 1; O .5 .6 .2'
    mol.basis = 'ccpvdz'
    mol.build()

    mf = GKS(mol)
    mf.xc = 'b3lyp'
    mf.kernel()

    dm = mf.init_guess_by_1e(mol)
    dm = dm + 0j
    nao = mol.nao_nr()
    numpy.random.seed(12)
    dm[:nao,nao:] = numpy.random.random((nao,nao)) * .1j
    dm[nao:,:nao] = dm[:nao,nao:].T.conj()
    mf.kernel(dm)
    mf.canonicalize(mf.mo_coeff, mf.mo_occ)
    mf.analyze()
    print(mf.spin_square())
    print(mf.e_tot - -76.2760115704274)
